/*  Copyright (C) 2015 Zoltán Nagy
 *  Created by Zoltán Nagy on 2/2/15.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __cxx_cont_tree_alloc_h__
#define __cxx_cont_tree_alloc_h__ 1

/*   std includes   */


/*   project headers   */
#include <cxx/memory-resource.h>
#include <cxx/containers/cont-tree-iterator.h>



namespace duct {
  namespace container {
    
    
    /*   binary tree node allocator    */
    template<typename _Tp, bool _Recycle>
    struct __tree_allocator
    {
      /*    aliases   */
      using node = __tree_node<_Tp>;
      using link = __tree_node_link;
      
      using allocator_type = pmr::memory_resource*;
      using __node_alloc = pmr::polymorphic_allocator<node>;
      using __link_alloc = pmr::polymorphic_allocator<link>;
      
      /*    constructors   */
      __tree_allocator() : _M_resource{pmr::get_default_resource()}, _M_junk{nullptr} {}
      explicit __tree_allocator(allocator_type a) : _M_resource{a ? a : pmr::get_default_resource()}, _M_junk{nullptr} {}
      
      /*   copy  is deleted   */
      __tree_allocator(const __tree_allocator&) = delete;
      __tree_allocator& operator=(const __tree_allocator&) = delete;
      
      /*    move   */
      __tree_allocator(__tree_allocator&& a) : __tree_allocator{}
      {
        std::swap(_M_junk, a._M_junk);
        std::swap(_M_resource, a._M_resource);
      }
      
      __tree_allocator& operator=(__tree_allocator&& a) {
        this->swap(a);
        return *this;
      }
      
      /*   destructor   */
      ~__tree_allocator() {
         burn_junk();
      }
      
      /*   returns a pointer to the resource object   */
      pmr::memory_resource * resource() const { return _M_resource;}
      
      /*   swapping tree allocators   */
      void swap(__tree_allocator& a)
      {
        if(&a != this ) {
          std::swap(_M_junk, a._M_junk);
          std::swap(_M_resource, a._M_resource);
        }
      }
      
      /*   create the root node   */
      __tree_node_link * create_root() {
        __link_alloc alloc{_M_resource};
        link *buf = alloc.allocate(1u);
        alloc.construct(buf);
        return buf;
      }
      
      /*   deallocating memory of a node  */
      void delete_root(link *ptr) {
        if(ptr) __link_alloc{_M_resource}.deallocate(ptr, 1u);
      }

      /*   this function allocates memory for a node   */
      link * allocate();
      link * allocate(_Tp&& v);
      link * allocate(const _Tp& v);
      template<typename ..._Args> link * allocate(_Args&& ...);
      
      /*   deallocating memory of a node  */
      void deallocate(link *ptr) {
        if(ptr) {
          ptr->~link();
          __node_alloc{_M_resource}.deallocate(static_cast<node *>(ptr), 1u);
        }
      }
      
      /*  destroying the node but keep the memory allocated   */
      void destroy(link *ptr) {
        if(ptr) __node_alloc{_M_resource}.destroy(std::addressof(static_cast<node *>(ptr)->data));
      }
      
      /*   clean the junk stack */
      void burn_junk()
      {
        /*   deallocating the junk stack   */
        link *it = _M_junk;
        
        while(it) {
          _M_junk = it->left;
          if(_Recycle) destroy(it);
          this->deallocate(it);
          it = _M_junk;
        }
      }
      
      /*   destroy the data if it is neded (do not deallocate) and move the nodes into the junk stack   */
      void dump_a_branch(link *);
      
      /*   deallocate the subtree   */
      void delete_a_branch(link *);
      
      /*   copy the subtree   */
      link * copy_a_branch(const link *);
      
      /*   move the subtree by nodes  */
      link * move_a_branch(link *);
      
      /*   append a node if it is a free slot  */
      link * append_left(link *);
      link * append_right(link *);
      
      link * append_left(link *, _Tp&&);
      link * append_right(link *, _Tp&&);
      
      link * append_left(link *, const _Tp&);
      link * append_right(link *, const _Tp&);
      
      template<typename ..._Args> link * emplace_left(link *, _Args&& ...);
      template<typename ..._Args> link * emplace_right(link *, _Args&& ...);
      
      /*   append a node if it is a free slot  */
      link * insert_left(link *);
      link * insert_right(link *);
      
      link * insert_left(link *, _Tp&&);
      link * insert_right(link *, _Tp&&);
      
      link * insert_left(link *, const _Tp&);
      link * insert_right(link *, const _Tp&);
      
      template<typename ..._Args> link * emplace_insert_left(link *, _Args&& ...);
      template<typename ..._Args> link * emplace_insert_right(link *, _Args&& ...);
      
    private:
      /*   underlying allocator   */
      allocator_type _M_resource;
      
      /*   garbage recycling   */
      link *_M_junk;
      
      /*   recycle a node    */
      link * pop()
      {
        if(_M_junk) {
          link *ret = _M_junk;
          
          _M_junk = ret->left;
          ret->left = ret->right = ret->up = nullptr;
          
          return ret;
        }
        
        return nullptr;
      }
    };
    
    
    
    template<typename _Tp, bool _Recycle>
    __tree_node_link * __tree_allocator<_Tp, _Recycle>::allocate()
    {
      __tree_node_link *c = pop();
      
      /*   successfully recycled node   */
      if(c && !_Recycle) {
        node *buf = static_cast<node *>(c);
        __node_alloc{_M_resource}.construct(std::addressof(buf->data));
      }
      
      /*   couldn't recycle   */
      else if(!c) {
        /*   allcating space for a node    */
        __node_alloc alloc{_M_resource};
        node *buf = alloc.allocate(1u);
        c = static_cast<link *>(buf);
        
        new(std::addressof(*c)) __tree_node_link{};
        alloc.construct(std::addressof(buf->data));
      }
      
      return c;
    }
    
    
    template<typename _Tp, bool _Recycle>
    __tree_node_link * __tree_allocator<_Tp, _Recycle>::allocate(_Tp&& v)
    {
      __tree_node_link *c = pop();
      
      /*   successfully recycled node   */
      if(c) {
        node *buf = static_cast<node *>(c);
        if(_Recycle) buf->data = std::forward<_Tp>(v);
        else __node_alloc{_M_resource}.construct(std::addressof(buf->data), std::forward<_Tp>(v));
      }
      
      /*   couldn't recycle   */
      else {
        /*   allcating space for a node    */
        __node_alloc alloc{_M_resource};
        node *buf = alloc.allocate(1u);
        c = static_cast<link *>(buf);
        
        new(std::addressof(*c)) __tree_node_link{};
        alloc.construct(std::addressof(buf->data), std::forward<_Tp>(v));
      }
      
      return c;
    }
    
    
    template<typename _Tp, bool _Recycle>
    __tree_node_link * __tree_allocator<_Tp, _Recycle>::allocate(const _Tp& v)
    {
      __tree_node_link *c = pop();
      
      /*   successfully recycled node   */
      if(c) {
        node *buf = static_cast<node *>(c);
        if(_Recycle) buf->data = v;
        else __node_alloc{_M_resource}.construct(std::addressof(buf->data), v);
      }
      
      /*   couldn't recycle   */
      else {
        /*   allcating space for a node    */
        __node_alloc alloc{_M_resource};
        node *buf = alloc.allocate(1u);
        c = static_cast<link *>(buf);
        
        new(std::addressof(*c)) __tree_node_link{};
        alloc.construct(std::addressof(buf->data), v);
      }
      
      return c;
    }
    
    
    template<typename _Tp, bool _Recycle> template<typename ..._Args>
    __tree_node_link * __tree_allocator<_Tp, _Recycle>::allocate(_Args&& ...args)
    {
      __tree_node_link *c = pop();
      
      /*   successfully recycled node   */
      if(c) {
        node *buf = static_cast<node *>(c);
        if(_Recycle) buf->data = _Tp{std::forward<_Args>(args)...};
        else __node_alloc{_M_resource}.construct(std::addressof(buf->data), std::forward<_Args>(args)...);
      }
      
      /*   couldn't recycle   */
      else {
        /*   allcating space for a node    */
        __node_alloc alloc{_M_resource};
        node *buf = alloc.allocate(1u);
        c = static_cast<link *>(buf);
        
        new(std::addressof(*c)) __tree_node_link{};
        alloc.construct(std::addressof(buf->data), std::forward<_Args>(args)...);
      }
      
      return c;
    }
    
    
    template<typename _Tp, bool _Recycle>
    void __tree_allocator<_Tp, _Recycle>::dump_a_branch(link *c)
    {
      /*   nothing to do   */
      if(c == nullptr) return;
      
      /*   separeting form the rest of the tree   */
      if(c->up) {
        if(c->up->left == c) c->up->left = nullptr;
        else c->up->right = nullptr;
      }
      
      /*   making c to head  */
      c->up = nullptr;
      
      /*   find the post-order begin    */
      while(c->left || c->right) c = (c->left ? c->left : c->right);
      
      /*   destroy the data and moving the node to the junk stack   */
      while(c) {
        link *todel = c;
        
        if(c->up == nullptr || c->up->right == c) c = c->up;
        else {
          c = c->up;
          if(c->right) {
            c = c->right;
            while(c->left || c->right)
              c = c->left ? c->left : c->right;
          }
        }
        
        /*   destroy the data if we don't recycle and link the node to the junk list   */
        if(!_Recycle) destroy(todel);
        todel->left = _M_junk;
        _M_junk = todel;
      }
    }
    
    
    template<typename _Tp, bool _Recycle>
    void __tree_allocator<_Tp, _Recycle>::delete_a_branch(link *c)
    {
      /*   nothing to do   */
      if(c == nullptr) return;
      
      /*   separeting form the rest of the tree   */
      if(c->up) {
        if(c->up->left == c) c->up->left = nullptr;
        else c->up->right = nullptr;
      }
      
      /*   making c to head  */
      c->up = nullptr;

      /*   find the post-order begin    */
      while(c->left || c->right) c = (c->left ? c->left : c->right);
      
      /*   destroy the data and moving the node to the junk stack   */
      while(c) {
        link *todel = c;
        
        if(c->up == nullptr || c->up->right == c) c = c->up;
        else {
          c = c->up;
          if(c->right) {
            c = c->right;
            while(c->left || c->right)
              c = c->left ? c->left : c->right;
          }
        }
        
        /*   destroy the data delete the node   */
        destroy(todel);
        deallocate(todel);
      }
    }
    
    
    template<typename _Tp, bool _Recycle>
    __tree_node_link * __tree_allocator<_Tp, _Recycle>::copy_a_branch(const link *xc)
    {
      /*   if x is empty, do nothing   */
      if(xc == nullptr) return nullptr;
      
      /*    copy the nodes    */
      link *end = xc->up;
      link *head, *u = nullptr;
      link **c = &head;
      
      while(xc != end) {
        *c = allocate(static_cast<const node *>(xc)->data);
        (*c)->up = u;
        u = *c;
        
        if(xc->left) {
          xc = xc->left;
          c = &(u->left);
        }
        else if(xc->right) {
          xc = xc->right;
          c = &(u->right);
        }
        else {
          while(xc->up != end && (xc->up->right == xc || xc->up->right == nullptr)) {
            xc = xc->up;
            u = u->up;
          }
          
          if(xc->up != end) {
            xc = xc->up->right;
            u = u->up;
            c = &(u->right);
          } else break;
        }
      }
      
      return head;
    }
    
    template<typename _Tp, bool _Recycle>
    __tree_node_link * __tree_allocator<_Tp, _Recycle>::move_a_branch(link *xc)
    {
      /*   if x is empty, do nothing   */
      if(xc == nullptr) return nullptr;
      
      /*    copy the nodes    */
      link *end = xc->up;
      link *head, *u = nullptr;
      link **c = &head;
      
      while(xc != end) {
        *c = allocate(std::move(static_cast<node *>(xc)->data));
        (*c)->up = u;
        u = *c;
        
        if(xc->left) {
          xc = xc->left;
          c = &(u->left);
        }
        else if(xc->right) {
          xc = xc->right;
          c = &(u->right);
        }
        else {
          while(xc->up != end && (xc->up->right == xc || xc->up->right == nullptr)) {
            xc = xc->up;
            u = u->up;
          }
          
          if(xc->up != end) {
            xc = xc->up->right;
            u = u->up;
            c = &(u->right);
          } else break;
        }
      }
      
      return head;
    }
    
  
    template<typename _Tp, bool _Recycle>
    inline __tree_node_link * __tree_allocator<_Tp, _Recycle>::
    append_left(__tree_node_link *n)
    {
      if(n->left == nullptr) {
        n->left = allocate();
        n->left->up = n;
        return n->left;
      }
      
      return n;
    }
    
    template<typename _Tp, bool _Recycle>
    inline __tree_node_link * __tree_allocator<_Tp, _Recycle>::
    append_right(__tree_node_link *n)
    {
      if(n->right == nullptr) {
        n->right = allocate();
        n->right->up = n;
        return n->right;
      }
      
      return n;
    }
    
    
    template<typename _Tp, bool _Recycle>
    inline __tree_node_link * __tree_allocator<_Tp, _Recycle>::
    append_left(__tree_node_link *n, _Tp&& d)
    {
      if(n->left == nullptr) {
        n->left = allocate(std::forward<_Tp>(d));
        n->left->up = n;
        return n->left;
      }
      
      return n;
    }
    
    template<typename _Tp, bool _Recycle>
    inline __tree_node_link * __tree_allocator<_Tp, _Recycle>::
    append_right(__tree_node_link *n, _Tp&& d)
    {
      if(n->right == nullptr) {
        n->right = allocate(std::forward<_Tp>(d));
        n->right->up = n;
        return n->right;
      }
      
      return n;
    }
    
    
    template<typename _Tp, bool _Recycle>
    inline __tree_node_link * __tree_allocator<_Tp, _Recycle>::
    append_left(__tree_node_link *n, const _Tp& d)
    {
      if(n->left == nullptr) {
        n->left = allocate(d);
        n->left->up = n;
        return n->left;
      }
      
      return n;
    }
    
    template<typename _Tp, bool _Recycle>
    inline __tree_node_link * __tree_allocator<_Tp, _Recycle>::
    append_right(__tree_node_link *n, const _Tp& d)
    {
      if(n->right == nullptr) {
        n->right = allocate(d);
        n->right->up = n;
        return n->right;
      }
      
      return n;
    }
    
    
    template<typename _Tp, bool _Recycle> template<typename ..._Args>
    inline __tree_node_link * __tree_allocator<_Tp, _Recycle>::
    emplace_left(__tree_node_link *n, _Args&& ...args)
    {
      if(n->left == nullptr) {
        n->left = allocate(std::forward<_Args>(args)...);
        n->left->up = n;
        return n->left;
      }
      
      return n;
    }
    
    template<typename _Tp, bool _Recycle> template<typename ..._Args>
    inline __tree_node_link * __tree_allocator<_Tp, _Recycle>::
    emplace_right(__tree_node_link *n, _Args&& ...args)
    {
      if(n->right == nullptr) {
        n->right = allocate(std::forward<_Args>(args)...);
        n->right->up = n;
        return n->right;
      }
      
      return n;
    }
    
    
    template<typename _Tp, bool _Recycle>
    inline __tree_node_link * __tree_allocator<_Tp, _Recycle>::
    insert_left(__tree_node_link *n)
    {
      if(n->up == nullptr) {
        __tree_node_link *i, *up = n->up;
        
        if(up->left == n) i = up->left = allocate();
        else i = up->right = allocate();
        
        up = i;
        i->left = n;
        return i;
      }
      
      return n;
    }
    
    template<typename _Tp, bool _Recycle>
    inline __tree_node_link * __tree_allocator<_Tp, _Recycle>::
    insert_right(__tree_node_link *n)
    {
      if(n->up == nullptr) {
        __tree_node_link *i, *up = n->up;
        
        if(up->left == n) i = up->left = allocate();
        else i = up->right = allocate();
        
        up = i;
        i->right = n;
        return i;
      }
      
      return n;
    }
    
    
   template<typename _Tp, bool _Recycle>
    inline __tree_node_link * __tree_allocator<_Tp, _Recycle>::
    insert_left(__tree_node_link *n, _Tp&& d)
    {
      if(n->up == nullptr) {
        __tree_node_link *i, *up = n->up;
        
        if(up->left == n) i = up->left = allocate(std::forward<_Tp>(d));
        else i = up->right = allocate(std::forward<_Tp>(d));
        
        up = i;
        i->left = n;
        return i;
      }
      
      return n;
    }
    
    template<typename _Tp, bool _Recycle>
    inline __tree_node_link * __tree_allocator<_Tp, _Recycle>::
    insert_right(__tree_node_link *n, _Tp&& d)
    {
      if(n->up == nullptr) {
        __tree_node_link *i, *up = n->up;
        
        if(up->left == n) i = up->left = allocate(std::forward<_Tp>(d));
        else i = up->right = allocate(std::forward<_Tp>(d));
        
        up = i;
        i->right = n;
        return i;
      }
      
      return n;
    }
    
    
    template<typename _Tp, bool _Recycle>
    inline __tree_node_link * __tree_allocator<_Tp, _Recycle>::
    insert_left(__tree_node_link *n, const _Tp& d)
    {
      if(n->up == nullptr) {
        __tree_node_link *i, *up = n->up;
        
        if(up->left == n) i = up->left = allocate(d);
        else i = up->right = allocate(d);
        
        up = i;
        i->left = n;
        return i;
      }
      
      return n;
    }
    
    template<typename _Tp, bool _Recycle>
    inline __tree_node_link * __tree_allocator<_Tp, _Recycle>::
    insert_right(__tree_node_link *n, const _Tp& d)
    {
      if(n->up == nullptr) {
        __tree_node_link *i, *up = n->up;
        
        if(up->left == n) i = up->left = allocate(d);
        else i = up->right = allocate(d);
        
        up = i;
        i->right = n;
        return i;
      }
      
      return n;
    }
    
    
    template<typename _Tp, bool _Recycle> template<typename ..._Args>
    inline __tree_node_link * __tree_allocator<_Tp, _Recycle>::
    emplace_insert_left(__tree_node_link *n, _Args&& ...args)
    {
      if(n->up == nullptr) {
        __tree_node_link *i, *up = n->up;
        
        if(up->left == n) i = up->left = allocate(std::forward<_Args>(args)...);
        else i = up->right = allocate(std::forward<_Args>(args)...);
        
        up = i;
        i->left = n;
        return i;
      }
      
      return n;
    }
    
    template<typename _Tp, bool _Recycle> template<typename ..._Args>
    inline __tree_node_link * __tree_allocator<_Tp, _Recycle>::
    emplace_insert_right(__tree_node_link *n, _Args&& ...args)
    {
      if(n->up == nullptr) {
        __tree_node_link *i, *up = n->up;
        
        if(up->left == n) i = up->left = allocate(std::forward<_Args>(args)...);
        else i = up->right = allocate(std::forward<_Args>(args)...);
        
        up = i;
        i->right = n;
        return i;
      }
      
      return n;
    }
    
  
  }    //  namespace containers
}    //  namespace duct

#endif
