/*  Copyright (C) 2015 Zoltán Nagy
 *  Created by Zoltán Nagy on 1/24/15.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

/*   std includes   */
#include <cerrno>
#include <cassert>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <utime.h>

/*   project headers  */
#include "fs-status.h"
#include "fs-error.h"
#include "fs-opers.h"
#include "fs-dir.h"
#include "fs-posix-api.h"



namespace duct {
  namespace filesystem {
   
    namespace api = __posix_api;

    /*   set permissions   */
    void permissions(const path& p, perms prms, std::error_code& ec) noexcept
    {
      /*  fresh start...   */
      ec.clear();
      
      /*   nothing to do both add_perms and remove perms bits are presented   */
      if((prms & perms::add_perms) != perms::none && (prms & perms::remove_perms) != perms::none)
        return;
      
      /*   get the current permissions    */
      auto curr = (static_cast<int>(prms & perms::resolve_symlinks)
                   ? symlink_status(p, ec).permissions()
                   : status(p, ec).permissions());
      
      /*   return if ther was an error to read the current permissions    */
      if(ec) return;
      
      /*  let's change them!   */
      int errval = 0;
      
      if((prms & perms::add_perms) != perms::none) prms |= curr;
      else if((prms & perms::remove_perms) != perms::none) prms = curr & ~prms;
        
      api::permissions(p, prms & perms::mask, errval);
      if(errval) ec = std::error_code(errval, std::generic_category());
    }
    
    
    /*   create directory  (mkdir)   */
    bool create_directory(const path& p, std::error_code& ec) noexcept
    {
      /*   fresh start  */
      ec.clear();
      
      /*   make an attempt for creating directory   */
      int errval = 0;
      bool res = api::create_directory(p, errval);
      
      if(errval) ec = std::error_code(errval, std::generic_category());
      
      return res;
    }
 
    bool create_directory(const path& p, const path& from, std::error_code& ec) noexcept
    {
      /*   fresh start  */
      ec.clear();
      
      /*   make an attempt for creating directory   */
      int errval = 0;
      bool res = api::create_directory(p, from, errval);
      
      if(errval) ec = std::error_code(errval, std::generic_category());
        
      return res;
    }
   
    
    /*   create directory  (mkdir -p)   */
    bool create_directories(const path& p, std::error_code& ec) noexcept
    {
      /*   fresh start  */
      ec.clear();
      
      /*  check if the directory is already exists   */
      std::error_code lec;
      
      if(is_directory(p, lec)) return false;
      
      /*   creating parent directory   */
      path parent = p.parent_path();
      
      /*   Internal error p == parent   */
      assert(parent != p);
      
      if(!parent.empty())
      {
        /*   if the parent does not exist, create the parent   */
        if(!exists(parent, lec)) {
          create_directories(parent, ec);
          if(ec) return false;
        }
      }
      
      /*   create the directory   */
      return create_directory(p, ec);
    }
    
    
    /*   create symlink   */
    void create_symlink(const path& to, const path& newlink, std::error_code& ec) noexcept
    {
      /*   fresh start  */
      ec.clear();
      
      /*   make an attempt for creating the link   */
      int errval = 0;
    
      api::create_symlink(to, newlink, errval);
      if(errval) ec = std::error_code(errval, std::generic_category());
    }
  
    
    /*   create directory symlink   */
    void create_directory_symlink(const path& to, const path& newlink, std::error_code& ec) noexcept
    {
      /*   fresh start  */
      ec.clear();
      
      /*   make an attempt for creating the link   */
      int errval = 0;
      
      api::create_directory_symlink(to, newlink, errval);
      if(errval) ec = std::error_code(errval, std::generic_category());
    }
    

    /*   create hard link   */
    void create_hard_link(const path& to, const path& newlink, std::error_code& ec) noexcept
    {
      /*   fresh start  */
      ec.clear();
      
      /*   make an attempt for creating the link   */
      int errval = 0;
      
      api::create_hard_link(to, newlink, errval);
      if(errval) ec = std::error_code(errval, std::generic_category());
    }
    
    
    /*   read current path  */
    path current_path(std::error_code& ec) noexcept
    {
      /*   fresh start  */
      ec.clear();
      
      /*   make an attempt for getting the current directory   */
      int errval = 0;
      auto res = api::current_path(errval);
      
      if(errval) ec = std::error_code(errval, std::generic_category());
        
      return res;
    }
    
    
    /*   set current path  */
    void current_path(const path& p, std::error_code& ec) noexcept
    {
      /*   fresh start  */
      ec.clear();
      
      /*   make an attempt for getting the current directory   */
      int errval = 0;
      
      api::current_path(p, errval);
      if(errval) ec = std::error_code(errval, std::generic_category());
    }

    
    /*   remove a file   */
    bool remove(const path& p, std::error_code& ec) noexcept
    {
      /*   fresh start  */
      ec.clear();

      /*  get the file type  */
      auto f = symlink_status(p, ec).type();
      
      if(ec || f == file_type::not_found) return false;
      
      /*   make an attempt for getting the current directory   */
      int errval = 0;
      auto res = api::remove(p, f, errval);
      
      if(errval) ec = std::error_code(errval, std::generic_category());
      return res;
    }

    
    /*   remove all (rm -r)   */
    std::uintmax_t remove_all(const path& p, std::error_code& ec) noexcept
    {
      /*   fresh start  */
      ec.clear();
      
      /*  get the file type  */
      auto type = symlink_status(p, ec).type();
      
      if(ec || type == file_type::not_found) return 0;
      
      /*   otherwise delete everything....   */
      std::error_code lec;
      std::uintmax_t count = 1;
      
      if(type == file_type::directory)
        for(const directory_entry& x : directory_iterator(p, ec)) {
          count += remove_all(x.path(), lec);
          if(lec) ec = lec;
        }
      
      /*   remove_dir_or_file  */
      remove(p, ec);
    
      return count;
    }
    
    
    /*   copy a file   */
    bool copy_file(const path& from, const path& to, copy_options option, std::error_code& ec) noexcept
    {
      /*   fresh start  */
      ec.clear();
      
      /*   check if destination exists   */
      bool to_exists = exists(to, ec);
      if(ec) return false;
      
      bool is_equiv = equivalent(from, to, ec);
      if(ec) return false;
      
      /*   report error and don't copy  */
      copy_options noreport = (copy_options::skip_existing | copy_options::overwrite_existing | copy_options::update_existing);
      
      if(to_exists && (is_equiv || (option & noreport) == copy_options::none)) {
        ec = std::error_code(EEXIST, std::system_category());
        return false;
      }
      
      /*   make copy  if `to` doesn't exist  */
      bool do_copy = !to_exists;
      
      /*  or `to` exists but we allow overwrite  */
      do_copy = do_copy || (to_exists && (option & copy_options::overwrite_existing) != copy_options::none);
      
      /*  or `to` exists but we allow update   */
      do_copy = do_copy || (to_exists && (option & copy_options::update_existing) != copy_options::none
                            && last_write_time(from, ec) > last_write_time(to, ec));
 
      if(ec) return false;
      
      /*   do the copy  */
      if(do_copy) {
        int errval = 0;

        /*  try to make the copy   */
        api::copy_file(from.c_str(), to.c_str(), errval);
        
        if(errval) {
          ec = std::error_code(errval, std::system_category());
          return false;
        }
        
        return true;
      }
      
      return false;
    }

    /*    copy everything...    */
    void copy(const path& from, const path& to, copy_options options, std::error_code& ec) noexcept
    {
      /*   fresh start  */
      ec.clear();
      
      /*   get the status    */
      bool is_symstat = ((options & copy_options::create_symlinks) != copy_options::none
                         || (options & copy_options::skip_symlinks) != copy_options::none);
      
      /*  get the status of the source file   */
      auto f = (is_symstat ? symlink_info(from, ec) : info(from, ec));
      if(ec) return;
      
      /*   get the status of the destinations   */
      auto t = (is_symstat ? symlink_info(to, ec) : info(to, ec));
      if(ec) return;
      
      /*   the copy is not possible    */
      bool to_notdir = is_directory(f) && is_regular_file(t);
      bool not_possible = !exists(f) || equivalent(from, to) || (is_other(f) || is_other(t)) || to_notdir;
      
      if(not_possible) {
        /*   source doesn't exists  */
        if(!exists(f)) ec = {ENOENT, std::system_category()};
        
        /*   sorce is a directory but destination is a file   */
        else if(to_notdir) ec = {ENOTDIR, std::system_category()};
        
        /*   the source and the destinations are equivalent   */
        else ec = {EPERM, std::system_category()};
        
        return;
      }

      /*   source is a symbolic link   */
      if(is_symlink(f)) {
        /*   skiping symlink   */
        if((options & copy_options::skip_symlinks) != copy_options::none) return;
        
        /*   copy symlinks if it is allowed   */
        if(!exists(t) && (options & copy_options::copy_symlinks) != copy_options::none)
          copy_symlink(from, to, ec);
        else ec = {EPERM, std::system_category()};
        
        return;
      }
 
      /*   source is a regular file   */
      if(is_regular_file(f)) {
        /*   we copy only the directory structure   */
        if((options & copy_options::directories_only) != copy_options::none) return;
        
        /*   create symbolic link instead of copying   */
        if((options & copy_options::create_symlinks) != copy_options::none)
          copy_symlink(from, to, ec);
        
        /*   create hard link instead of copying   */
        else if((options & copy_options::create_hard_links) != copy_options::none)
          create_hard_link(from, to, ec);
        
        /*   copy the file into a directory   */
        else if(is_directory(t)) copy_file(from, to/from.filename(), options, ec);
        
        /*   otherwise just copy it   */
        else copy_file(from, to, options, ec);
        
        return;
      }
 
      /*   source is a directory   */
      if(is_directory(f) && ((options & copy_options::recursive) != copy_options::none || options == copy_options::none)) {
        /*   create the destination directory if it is neded   */
        if(!exists(t)) create_directory(to, from, ec);
        if(ec) return;
        
        /*   then, iterate over the files in `from`   */
        std::error_code lec;
        
        for(const directory_entry& x : directory_iterator(from, ec)) {
          copy(x.path(), to/x.path().filename(), options, lec);
          if(lec) ec = lec;
        }
      }
      
      return;
    }
    

  }  //  namespace filesystem
}  //  namespace duct






#if 0

#   if !defined(__APPLE__) && !defined(__OpenBSD__) && !defined(__ANDROID__)
#     include <sys/statvfs.h>
#     define BOOST_STATVFS statvfs
#     define BOOST_STATVFS_F_FRSIZE vfs.f_frsize
#   else
#     ifdef __OpenBSD__
#     include <sys/param.h>
#     elif defined(__ANDROID__)
#     include <sys/vfs.h>
#     endif
#     include <sys/mount.h>
#     define BOOST_STATVFS statfs
#     define BOOST_STATVFS_F_FRSIZE static_cast<boost::uintmax_t>(vfs.f_bsize)
#   endif

#   define BOOST_MOVE_FILE(OLD,NEW)(::rename(OLD, NEW)== 0)
#   define BOOST_RESIZE_FILE(P,SZ)(::truncate(P, SZ)== 0)


namespace
{
  
  bool is_empty_directory(const path& p)
  {
    return fs::directory_iterator(p)== end_dir_itr;
  }
 
  bool not_found_error(int errval)
  {
    return errno == ENOENT || errno == ENOTDIR;
  }
  
  
} // unnamed namespace

namespace boost
{
  namespace filesystem
  {
    
    
    path absolute(const path& p, const path& base)
    {
      //    if ( p.empty() || p.is_absolute() )
      //      return p;
      //    //  recursively calling absolute is sub-optimal, but is simple
      //    path abs_base(base.is_absolute() ? base : absolute(base));
      //# ifdef BOOST_WINDOWS_API
      //    if (p.has_root_directory())
      //      return abs_base.root_name() / p;
      //    //  !p.has_root_directory
      //    if (p.has_root_name())
      //      return p.root_name()
      //        / abs_base.root_directory() / abs_base.relative_path() / p.relative_path();
      //    //  !p.has_root_name()
      //# endif
      //    return abs_base / p;
      
      //  recursively calling absolute is sub-optimal, but is sure and simple
      path abs_base(base.is_absolute() ? base : absolute(base));
      
      //  store expensive to compute values that are needed multiple times
      path p_root_name (p.root_name());
      path base_root_name (abs_base.root_name());
      path p_root_directory (p.root_directory());
      
      if (p.empty())
        return abs_base;
      
      if (!p_root_name.empty())  // p.has_root_name()
      {
        if (p_root_directory.empty())  // !p.has_root_directory()
          return p_root_name / abs_base.root_directory()
          / abs_base.relative_path() / p.relative_path();
        // p is absolute, so fall through to return p at end of block
      }
      
      else if (!p_root_directory.empty())  // p.has_root_directory()
      {
        // POSIX can have root name it it is a network path
        if (base_root_name.empty())   // !abs_base.has_root_name()
          return p;
        return base_root_name / p;
      }
      
      else
      {
        return abs_base / p;
      }
      
      return p;  // p.is_absolute() is true
    }
    
    namespace detail
    {
       bool possible_large_file_size_support()
      {
        struct stat lcl_stat;
        return sizeof(lcl_stat.st_size)> 4;
      }
      
      
      path canonical(const path& p, const path& base, system::error_code* ec)
      {
        path source (p.is_absolute() ? p : absolute(p, base));
        path root(source.root_path());
        path result;
        
        system::error_code local_ec;
        file_status stat (status(source, local_ec));
        
        if (stat.type() == fs::file_not_found)
        {
          if (ec == 0)
            BOOST_FILESYSTEM_THROW(filesystem_error(
                                                    "boost::filesystem::canonical", source,
                                                    error_code(system::errc::no_such_file_or_directory, system::generic_category())));
          ec->assign(system::errc::no_such_file_or_directory, system::generic_category());
          return result;
        }
        else if (local_ec)
        {
          if (ec == 0)
            BOOST_FILESYSTEM_THROW(filesystem_error(
                                                    "boost::filesystem::canonical", source, local_ec));
          *ec = local_ec;
          return result;
        }
        
        bool scan (true);
        while (scan)
        {
          scan = false;
          result.clear();
          for (path::iterator itr = source.begin(); itr != source.end(); ++itr)
          {
            if (*itr == dot_path())
              continue;
            if (*itr == dot_dot_path())
            {
              if (result != root)
                result.remove_filename();
              continue;
            }
            
            result /= *itr;
            
            bool is_sym (is_symlink(detail::symlink_status(result, ec)));
            if (ec && *ec)
              return path();
            
            if (is_sym)
            {
              path link(detail::read_symlink(result, ec));
              if (ec && *ec)
                return path();
              result.remove_filename();
              
              if (link.is_absolute())
              {
                for (++itr; itr != source.end(); ++itr)
                  link /= *itr;
                source = link;
              }
              else // link is relative
              {
                path new_source(result);
                new_source /= link;
                for (++itr; itr != source.end(); ++itr)
                  new_source /= *itr;
                source = new_source;
              }
              scan = true;  // symlink causes scan to be restarted
              break;
            }
          }
        }
        if (ec != 0)
          ec->clear();
        BOOST_ASSERT_MSG(result.is_absolute(), "canonical() implementation error; please report");
        return result;
      }
      
      
      
      path initial_path(error_code* ec)
      {
        static path init_path;
        if (init_path.empty())
          init_path = current_path(ec);
        else if (ec != 0) ec->clear();
        return init_path;
      }
      
      
      bool is_empty(const path& p, system::error_code* ec)
      {
        struct stat path_stat;
        if (error(::stat(p.c_str(), &path_stat)!= 0,
                  p, ec, "boost::filesystem::is_empty"))
          return false;
        return S_ISDIR(path_stat.st_mode)
        ? is_empty_directory(p)
        : path_stat.st_size == 0;
      }
      
      
      void rename(const path& old_p, const path& new_p, error_code* ec)
      {
        error(!BOOST_MOVE_FILE(old_p.c_str(), new_p.c_str()), old_p, new_p, ec,
              "boost::filesystem::rename");
      }
      
      
      void resize_file(const path& p, uintmax_t size, system::error_code* ec)
      {
        error(!BOOST_RESIZE_FILE(p.c_str(), size), p, ec, "boost::filesystem::resize_file");
      }
      
      
      space_info space(const path& p, error_code* ec)
      {
        struct BOOST_STATVFS vfs;
        space_info info;
        if (!error(::BOOST_STATVFS(p.c_str(), &vfs)!= 0,
                   p, ec, "boost::filesystem::space"))
        {
          info.capacity
          = static_cast<boost::uintmax_t>(vfs.f_blocks)* BOOST_STATVFS_F_FRSIZE;
          info.free
          = static_cast<boost::uintmax_t>(vfs.f_bfree)* BOOST_STATVFS_F_FRSIZE;
          info.available
          = static_cast<boost::uintmax_t>(vfs.f_bavail)* BOOST_STATVFS_F_FRSIZE;
        }
        
        
        else
        {
          info.capacity = info.free = info.available = 0;
        }
        return info;
      }
      
      
      
        // contributed by Jeff Flinn
      
      path temp_directory_path(system::error_code* ec)
      {
        const char* val = 0;
        
        (val = std::getenv("TMPDIR" )) ||
        (val = std::getenv("TMP"    )) ||
        (val = std::getenv("TEMP"   )) ||
        (val = std::getenv("TEMPDIR"));
        
        path p((val!=0) ? val : "/tmp");
        
        if (p.empty() || (ec&&!is_directory(p, *ec))||(!ec&&!is_directory(p)))
        {
          errno = ENOTDIR;
          error(true, p, ec, "boost::filesystem::temp_directory_path");
          return p;
        }
        
        return p;
      }
      
      
      path system_complete(const path& p, system::error_code* ec)
      {
        return (p.empty() || p.is_absolute())
        ? p : current_path()/ p;
      }
      
    }  // namespace detail
    
  
 
  } // namespace filesystem
} // namespace boost


#endif