/*  Copyright (C) 2015 Zoltán Nagy
 *  Created by Zoltán Nagy on 1/23/15.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

/*   std includes   */
#include <cassert>
#include <cstring>

/*   project headers  */
#include "fs-path.h"


namespace duct {
  namespace filesystem {
    
    
    path::path(const char *f)
    {
      /*   check if it is an absolute path   */
      bool skip, p_root = (f[0] =='/');
      
      if(p_root) {
        _M_path += '/';
        ++f;
      }
      
      /*  parse the whole string   */
      const char  *n;
      
      while((n = std::strchr(f, '/'))) {
        skip = ((f[0] =='.' && f[1] == '/')                                 //  f == "./"
                || f[0] =='/'                                               //  f == "/"
                || (p_root && f[0] == '.' && f[1] == '.' && f[2] == '/')    //  root (/) followed by f == ../
                );
        
        if(!skip) {
          p_root = false;
          _M_path.append(f, n+1);
        }
        
        f = n+1;
      }
      
      /*   appending the filename   */
      skip = ((!_M_path.empty() && std::strcmp(f,".") == 0)
              || std::strcmp(f,"/") == 0
              || (std::strcmp(f,"..") == 0 && p_root)
              );
      
      /*  append the filename if it is "real"   */
      if(!skip) _M_path.append(f);
      
      /*  removing trailing slahs   */
      if(_M_path.size() > 1 && _M_path.back() == '/') _M_path.pop_back();
    }
    
    
    path& path::operator/=(const path& p)
    {
      if(p.empty()) return *this;
      
      if(this == &p)  // self-append
      {
        auto rhs = _M_path;
        if(rhs[0] != '/' && rhs.back() != '/') _M_path += '/';
        _M_path += rhs;
      } else {
        if(p._M_path[0] != '/' && _M_path.back() != '/') _M_path += '/';
        _M_path += p._M_path;
      }
      
      return *this;
    }
    
    
    path path::filename() const
    {
      /*   the position of the last '/'   */
      auto pos = _M_path.find_last_of("/");
      
      /*   the path is a filename   */
      if(pos == std::string::npos) return *this;
      
      /*   otherwise copy out the filename   */
      path res;
      res._M_path = _M_path.c_str() + pos+1;
      
      return res;
    }
    
    
    path path::extension() const
    {
      path name(filename());
      if(name._M_path == "." || name._M_path == "..") return path();
      
      auto pos = name._M_path.rfind('.');
      
      return (pos == string_type::npos
              ? path()
              : path(name._M_path.c_str() + pos));
    }
    
    path& path::replace_extension(const path& ext)
    {
      path name(filename());
      if(name._M_path == "." || name._M_path == "..") return *this;
      
      auto pos = _M_path.rfind('.');
      
      if(pos != string_type::npos)
      {
        if(ext.empty()) {
          _M_path.erase(pos);
        } else {
          _M_path.erase(pos+1);
          _M_path += ext.string();
        }
      }
      
      return *this;
    }
    
    
    path path::stem() const
    {
      path name(filename());
      if(name._M_path == "." || name._M_path == "..") return name;
      
      auto pos = name._M_path.rfind('.');
      
      return (pos == string_type::npos
              ? name : path{name._M_path.c_str(), name._M_path.c_str() + pos});
    }
    
    
    path path::parent_path() const
    {
      const char *o = _M_path.c_str();
      const char  *n, *p = o, *f = o;
      
      while((n = std::strchr(f, '/'))) {
        p = f;
        f = n+1;
      }
      /*   it is a file name  */
      if(f == o)  return path{};
      
      /*  the parent is the root   */
      if(f-1 == o) return "/";
      
      /*   otherwise   */
      return *f == '\0' ? path{o,p-1} : path{o,f-1};
    }
    
    
    path::iterator& path::iterator::operator++()
    {
      /*   increment  */
      _M_curr = _M_next;
      
      /*   found the next   */
      auto next = _M_path->_M_path.find_first_of('/', _M_curr);
      _M_next = (next != std::string::npos ? next+1 : _M_path->_M_path.size());
      
      auto b = _M_path->_M_path.begin();
      _M_elem.assign(b + _M_curr, b + _M_next);
      
      return *this;
    }
    
    path::iterator& path::iterator::operator--()
    {
      /*   first element   */
      if(_M_curr == 0) {
        _M_curr = _M_next = 0;
        _M_elem = path{};
        return *this;
      }
      
      /*   we have root and this is right after root  */
      if(_M_curr == 1) {
        _M_curr = 0;
        _M_next = 1;
        _M_elem = path{"/"};
        return *this;
      }
      
      /*  otherwise increment  */
      _M_next = _M_curr;
      
      /*   found the previous   */
      auto prev = _M_path->_M_path.find_last_of('/', _M_path->_M_path[_M_next-1] == '/' ? _M_next-2 : _M_next-1);
      
      _M_curr = (prev == std::string::npos ? 0 : prev+1);
      
      auto b = _M_path->_M_path.begin();
      _M_elem.assign(b + _M_curr, b + _M_next);
      
      return *this;
    }

  }
}

