/*  Copyright (C) 2015 Zoltán Nagy
 *  Created by Zoltán Nagy on 1/24/15.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __cxx_fs_status_h__
#define __cxx_fs_status_h__ 1

/*   std includes   */
#include <cstdint>
#include <chrono>

/*   project headers   */
#include <cxx/filesystem/fs-path.h>
#include <cxx/filesystem/fs-error.h>



namespace duct {
  namespace filesystem {
    
    /*   Enum class for specifying the file type   */
    enum class file_type {
      not_found = -1,
      none,
      regular,
      directory,
      symlink,
      block,
      character,
      fifo,
      socket,
      unknown
      //,_detail_directory_symlink  // internal use only; never exposed to users
    };
    
    
    /*   Enum class for specifying the file permissions   */
    enum class perms {
      none = 0,             // no permission set
      
      owner_read  = 0400,   // S_IRUSR, Read permission, owner
      owner_write = 0200,   // S_IWUSR, Write permission, owner
      owner_exe   = 0100,   // S_IXUSR, Execute/search permission, owner
      owner_all   = 0700,   // S_IRWXU, Read, write, execute/search by owner
      
      group_read  = 040,    // S_IRGRP, Read permission, group
      group_write = 020,    // S_IWGRP, Write permission, group
      group_exe   = 010,    // S_IXGRP, Execute/search permission, group
      group_all   = 070,    // S_IRWXG, Read, write, execute/search by group
      
      others_read  = 04,    // S_IROTH, Read permission, others
      others_write = 02,    // S_IWOTH, Write permission, others
      others_exe   = 01,    // S_IXOTH, Execute/search permission, others
      others_all   = 07,    // S_IRWXO, Read, write, execute/search by others
      
      all          = 0777,  // owner_all|group_all|others_all
      
      set_uid      = 04000, // S_ISUID, Set-user-ID on execution
      set_gid      = 02000, // S_ISGID, Set-group-ID on execution
      sticky_bit   = 01000, // S_ISVTX,
     
      mask         = 07777,  // all_all | set_uid | set_gid | sticky_bit
      unknown      = 0xFFFF, // present when directory_entry cache not loaded

      // options for permissions() function
      add_perms    = 0x10000,     // adds the given permission bits to the current bits
      remove_perms = 0x20000,     // removes the given permission bits from the current bits;
      resolve_symlinks = 0x40000  // on POSIX, don't resolve symlinks; implied on Windows
    };
    
    /*    permissions are bitmask constants   */
    constexpr perms operator|(perms x , perms y) {
      return static_cast<perms>(static_cast<int>(x) | static_cast<int>(y));
    }
    
    constexpr perms operator&(perms x , perms y) {
      return static_cast<perms>(static_cast<int>(x) & static_cast<int>(y));
    }
    
    constexpr perms operator^(perms x , perms y) {
      return static_cast<perms>(static_cast<int>(x) ^ static_cast<int>(y)); }
    
    constexpr perms operator~(perms x) {
      return static_cast<perms>(~static_cast<int>(x));
    }
  
    constexpr perms& operator&=(perms& x , perms y) {
      x = x & y;
      return x;
    }
    
    constexpr perms& operator|=(perms& x , perms y) {
      x = x | y;
      return x;
    }
    
    constexpr perms& operator^=(perms & x , perms y) {
      x = x ^ y;
      return x;
    }
    
    /*   time type for file system   */
    using file_time_type = std::chrono::time_point<std::chrono::system_clock>;

    
    /*    simple class for representing the file status   */
    class file_status
    {
    public:
      /*   constructors   */
      explicit file_status(file_type v = file_type::none, perms prms = perms::unknown)
      : _M_value(v), _M_perms(prms) {}
      
      /*   observers    */
      file_type type() const { return _M_value;}
      perms permissions() const { return _M_perms;}
      
      /*    modifiers   */
      void type(file_type v) { _M_value = v; }
      void permissions(perms prms) { _M_perms = prms; }
      
      /*   comparison   */
      bool operator==(const file_status& rhs) const {
        return type() == rhs.type() && permissions() == rhs.permissions();
      }
      
      bool operator!=(const file_status& rhs) const {
        return !(*this == rhs);
      }
      
    private:
      /*   data members   */
      file_type _M_value;
      perms _M_perms;
    };
   
 
    /*   structure for space info   */
    struct space_info
    {
      // all values are byte counts
      std::uintmax_t capacity;
      std::uintmax_t free;      // <= capacity
      std::uintmax_t available; // <= free
    };
    
 
    
    /*   This class gives more information on a file.  This is not part of the standard 
     *   but help to implement the standard in a simpler way...
     */
    class file_info : public file_status
    {
    private:
      /*   it is default constructable for the privilaged functions   */
      file_info() : _M_stat(nullptr) {};
      
    public:
      /*   it is not copyable   */
      file_info(const file_info&) = delete;
      file_info& operator=(const file_info&) = delete;
      
      /*   only movable   */
      file_info(file_info&&);
      file_info& operator=(file_info&&);
      
      /*  destructor  */
      ~file_info() { deallocate();}
      
      /*  check if the files are equivalent   */
      bool equivalent(const file_info&) const noexcept;
      
      /*   observers    */
      file_type type() const  noexcept{ return file_status::type();}
      perms permissions() const  noexcept{ return file_status::permissions();}
      
      /*   observers: these are not standard but it is logical to include them   */
      std::uintmax_t hard_link_count() const noexcept;
      std::uintmax_t file_size() const noexcept;
      file_time_type last_write_time() const noexcept;
      
      /*   make info   */
      static file_info info(const path&, int&) noexcept;
      static file_info symlink_info(const path&, int&) noexcept;
      
    private:
      /*   the underlying object is platform dependent  */
      void *_M_stat;

      /*   allocate/deallocate memory   */
      int allocate() noexcept;
      void deallocate() noexcept;
    };
    
    
    /*   make file info   */
    inline file_info info(const path& p)
    {
      int errval = 0;
      auto retval = file_info::info(p, errval);
      
      if(errval) throw filesystem_error("duct::filesystem::info(...)", p, std::error_code(errval, std::system_category()));
      
      return retval;
    }
    
    inline file_info info(const path& p, std::error_code& ec) noexcept
    {
      int errval = 0;
      auto retval = file_info::info(p, errval);
      ec = std::error_code(errval, std::system_category());
      
      return retval;
    }
    
    inline file_info symlink_info(const path& p)
    {
      int errval = 0;
      auto retval = file_info::symlink_info(p, errval);
      
      if(errval) throw filesystem_error("duct::filesystem::symlink_info(...)", p, std::error_code(errval, std::system_category()));
      
      return retval;
    }
    
    inline file_info symlink_info(const path& p, std::error_code& ec) noexcept
    {
      int errval = 0;
      auto retval = file_info::symlink_info(p, errval);
      ec = std::error_code(errval, std::system_category());
      
      return retval;
    }
    
    
    /*   status query  functions   */
    inline file_status status(const path& p) {
      return info(p);
    }
    
    inline file_status status(const path& p, std::error_code& ec) noexcept {
      return info(p, ec);
    }
    
    inline file_status symlink_status(const path& p) {
      return symlink_info(p);
    }
    
    inline file_status symlink_status(const path& p, std::error_code& ec) noexcept {
      return symlink_info(p, ec);
    }
    
    /*   returns true if the files are equivalent   */
    inline bool equivalent(const file_info& p1, const file_info& p2) noexcept {
      return p1.equivalent(p2);
    }
    
    inline bool equivalent(const path& p1, const path& p2, std::error_code& ec) noexcept
    {
      ec.clear();
      
      auto f1 = info(p1, ec);
      if(ec) return false;
   
      auto f2 = info(p2, ec);
      if(ec) return false;
      
      return f1.equivalent(f2);
    }
  
    inline bool equivalent(const path& p1, const path& p2)
    {
      std::error_code ec;
      auto res = equivalent(p1, p2, ec);
      if(ec) throw filesystem_error("duct::filesystem::equivalent(...)", p1, p2, ec);
      return res;
    }
    
   
    /*   read the linked filename   */
    path read_symlink(const path&, std::error_code&) noexcept;
    
    inline path read_symlink(const path& p)
    {
      std::error_code ec;
      auto res = read_symlink(p, ec);
      if(ec) throw filesystem_error("duct::filesystem::read_symlink(...)", p, ec);
      return res;
    }

    /*    change the last write time   */
    void last_write_time(const path&, const file_time_type, std::error_code&) noexcept;

    inline void last_write_time(const path& p, const file_time_type new_time)
    {
      std::error_code ec;
      last_write_time(p, new_time, ec);
      if(ec) throw filesystem_error("duct::filesystem::last_write_time(...)", p, ec);
    }

    
    /*   returns true if the status is known      */
    inline bool status_known(const file_status& f) noexcept {
      return (f.type() != file_type::none && f.permissions() != perms::unknown);
    }
    
    
    /*   returns true if the file exists      */
    inline bool exists(const file_status& f) noexcept {
      return f.type() != file_type::none && f.type() != file_type::not_found;
    }
    
    inline bool exists(const path& p) {
      return exists(status(p));
    }
    
    inline bool exists(const path& p, std::error_code& ec) noexcept {
      return exists(status(p, ec));
    }

    
    /*   returns the file size       */
   inline std::uintmax_t file_size(const file_info& f) noexcept {
      return f.file_size();
    }
    
    inline std::uintmax_t file_size(const path& p) {
      return file_size(info(p));
    }
    
    inline std::uintmax_t file_size(const path& p, std::error_code& ec) noexcept {
      return file_size(info(p, ec));
    }

    
    /*   returns hard link count      */
    inline std::uintmax_t hard_link_count(const file_info& f) noexcept {
      return f.hard_link_count();
    }
  
    inline std::uintmax_t hard_link_count(const path& p) {
      return hard_link_count(info(p));
    }
    
    inline std::uintmax_t hard_link_count(const path& p, std::error_code& ec) noexcept {
      return hard_link_count(info(p, ec));
    }
    
    
    /*   returns true if it is block file       */
    inline bool is_block_file(const file_status& f) noexcept {
      return f.type() == file_type::block;
    }

    inline bool is_block_file(const path& p) {
      return is_block_file(status(p));
    }
    
    inline bool is_block_file(const path& p, std::error_code& ec) noexcept {
      return is_block_file(status(p, ec));
    }
    
    
    /*   returns true if it is character file       */
   inline bool is_character_file(file_status f) noexcept {
      return f.type() == file_type::character;
    }
    
    inline bool is_character_file(const path& p) {
      return is_character_file(status(p));
    }
    
    inline bool is_character_file(const path& p, std::error_code& ec) noexcept {
      return is_character_file(status(p, ec));
    }
    
    
    /*   returns true if is directory       */
    inline bool is_directory(const file_status& f) noexcept {
      return f.type() == file_type::directory;
    }

    inline bool is_directory(const path& p) {
      return is_directory(status(p));
    }
    
    inline bool is_directory(const path& p, std::error_code& ec) noexcept {
      return is_directory(status(p, ec));
    }
    
    
    /*   returns true if it is fifo       */
    inline bool is_fifo(const file_status& f) noexcept {
      return f.type() == file_type::fifo;
    }
    
    inline bool is_fifo(const path& p) {
      return is_fifo(status(p));
    }
    
    inline bool is_fifo(const path& p, std::error_code& ec) noexcept {
      return is_fifo(status(p, ec));
    }

    
    /*   returns true if it is regular file       */
    inline bool is_regular_file(const file_status& f) noexcept {
      return f.type() == file_type::regular;
    }
    
    inline bool is_regular_file(const path& p) {
      return is_regular_file(status(p));
    }
    
    inline bool is_regular_file(const path& p, std::error_code& ec) noexcept {
      return is_regular_file(status(p, ec));
    }
    
    
    /*   returns true if it is socket       */
    inline bool is_socket(const file_status& f) noexcept {
      return f.type() == file_type::socket;
    }
    
    inline bool is_socket(const path& p) {
      return is_socket(status(p));
    }
    
    inline bool is_socket(const path& p, std::error_code& ec) noexcept {
      return is_socket(status(p, ec));
    }

    
    /*   returns true if it is symbolic link       */
    inline bool is_symlink(const file_status& f) noexcept {
      return f.type() == file_type::symlink;
    }
    
    inline bool is_symlink(const path& p) {
      return is_symlink(symlink_status(p));
    }
    
    inline bool is_symlink(const path& p, std::error_code& ec) noexcept {
      return is_symlink(symlink_status(p, ec));
    }
    
    
    /*   returns true if it is not regular diractory or symlink       */
    inline bool is_other(const file_status& f) noexcept {
      return exists(f) && !is_regular_file(f) && !is_directory(f) && !is_symlink(f);
    }
    
    inline bool is_other(const path& p) {
      return is_other(status(p));
    }
    
    inline bool is_other(const path& p, std::error_code& ec) noexcept {
      return is_other(status(p, ec));
    }
    

    /*   returns the last write time      */
    inline file_time_type last_write_time(const file_info& f) {
      return f.last_write_time();
    }
 
    inline file_time_type last_write_time(const path& p) {
      return last_write_time(info(p));
    }
    
    inline file_time_type last_write_time(const path& p, std::error_code& ec) noexcept {
      return last_write_time(info(p, ec));
    }

  }
}

#endif
