/*  Copyright © 2015 Zoltán Nagy. All rights reserved.
 *  Created by Zoltán Nagy on 10/12/15.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __math_spline_h__
#define __math_spline_h__ 1


/*   std includes    */
#include <cmath>
#include <limits>
#include <array>
#include <vector>
#include <algorithm>


/*   project headers   */
#include <cxx/math/lambert.h>



namespace duct {

  
  /*   enum class for parametrizing the order of the spline   */
  enum class spline_type {
    constant = 0,
    linear
  };
  
  
  /*    This is the definition of unit weight for splines.    */
  struct unitweight
  {
    template<spline_type _Order>
    using _Poly = std::array<double, 1+static_cast<std::size_t>(_Order)>;
    
    constexpr double operator()(double) const {
      return 1.0;
    }
    
    template<spline_type _Order>
    double integral(const _Poly<_Order>& p, double x1, double x2) const
    {
      if(spline_type::constant == _Order) return p[0]*(x2-x1);
      else return (p[0] + 0.5*p[1]*(x1+x2))*(x2-x1);
    }
    
    template<spline_type _Order>
    double integral_inverse(const _Poly<_Order>& p, double x1, double x2, double F) const
    {
      if(spline_type::constant == _Order)
        return x1 + F/p[0];
      else {
        double c = F + p[0]*x1 + 0.5*p[1]*x1*x1;
        double d = std::sqrt(p[0]*p[0] + 2.0*p[1]*c);
        return 2.0*c/(p[0] + d);
      }
    }
  };
  
  
  
  /*   The weight function that leads to logarithm after integration.
   *   Thus we define
   *
   *             /       1
   *             |    --------    for x <= xT
   *     w(x) =  |     xT - x
   *             |
   *             \       0        x > xT
   *
   */
  struct logweight
  {
  private:
    /*   simple aliass that we always need   */
    template<spline_type _Order>
    using _Poly = std::array<double, 1+static_cast<std::size_t>(_Order)>;
    
    /*   the point od the singularity    */
    double xT;
    
  public:
    /*   constructor   */
    explicit logweight(double x) : xT{x} {}
    
    /*   weight function    */
    double operator()(double x) const {
      return (x > xT ? 0.0 : 1.0/std::max(xT-x, 2.0*std::numeric_limits<double>::min()));
    }
    
    /*   its integral over a spline    */
    template<spline_type _Order>
    double integral(const _Poly<_Order>& p, double x1, double x2) const
    {
      /*   trivial case   */
      if(x1 > xT) return 0.0;
      
      /*  calculate the integral   */
      double l12 = std::log((xT-x1)/std::max(xT-x2, 2.0*std::numeric_limits<double>::min()));
      
      if(spline_type::constant == _Order) return p[0]*l12;
      else return (p[0] + xT*p[1])*l12 - p[1]*(std::min(x2, xT)-x1);
    }
    
    /*    inverse of the integral    */
    template<spline_type _Order>
    double integral_inverse(const _Poly<_Order>& p, double x1, double x2, double F) const
    {
      static constexpr double tiny = 2.0*std::numeric_limits<double>::min();
      
      if(spline_type::constant == _Order)
        return xT - (xT-x1)*std::exp(-F/p[0]);
      else {
        /*   adjusting the upper bound   */
        if(x2 > xT) x2 = xT;
        
        /*   calculate the primitiv functions at the boundaries   */
        double A = p[0]+xT*p[1], B = p[1];
        double xTx = std::max(xT-x2, tiny);
        double F1 = B*(xT-x1) - A*std::log(xT-x1);
        
        /*    try to find the root numerically   */
        double I, dI, ox, x = 0.5*(x1 + x2);
        
        F += F1;
        
        do {
          ox = x;
          xTx = std::max(xT-x, tiny);
          I = B*xTx - A*std::log(xTx) - F;
          
          if(I <= 0.0) x1 = x;
          if(I >= 0.0) x2 = x;
          
          /*    start with the Newton method   */
          dI = A/xTx-B;
          x -= I/dI;
          
          /*   if it fails we go with the bisect method   */
          if(x < x1 || x > x2) x = 0.5*(x1+x2);
        } while(std::abs(x-ox) > 1e-8*std::abs(x1+x2));
        
        return x;
      }
    }
  };
  
  
  
  /*    Simple spline implementation   */
  template<spline_type _Order>
  class spline
  {
  public:
    /*    simple grid point type   */
    struct gridpoint_t {
      double x, func, ifunc;
    };
    
    /*   order of the spaline as static member    */
    static constexpr spline_type order = _Order;
    
    /*   constructors     */
    spline() = default;
    explicit spline(std::size_t n,  double xL = 0.0, double xH = 1.0) { makegrid(n, xL, xH);}
    spline(std::size_t n, std::initializer_list<double> xB) { makegrid(n, xB);}
    
    /*   making grid    */
    void makegrid(std::size_t, double, double);
    void makegrid(std::size_t, std::initializer_list<double>);
    
    
    /*   initialize the function  (no integral initialization)   */
    template<class _Func> void assign(_Func&& func) {
      for(gridpoint_t& g : _M_grid) g.func = func(g.x);
    }
    
    /*   initialize the integral     */
    double assign_integral() { return assign_integral(unitweight{});}
    
    
    /*    do the interpolation with distribution   */
    double operator()(double x) const { return operator()(unitweight{}, x);}
    
    /*    interpolate the integral this assumes a pre-call of integral_init()    */
    double integral(double x) const { return integral(unitweight{}, x);}
    
    /*    integral inverse this assumes a pre-call of integral_init()    */
    double integral_inverse(double F) const { return integral_inverse(unitweight{}, F);}
    
    
    /*    element access    */
    gridpoint_t& operator[](std::size_t k) { return _M_grid[k];}
    const gridpoint_t& operator[](std::size_t k) const { return _M_grid[k];}
    
    /*   observers   */
    std::size_t size() const { return _M_grid.size();}
    double integral() const { return _M_grid.back().ifunc;}
    
    /*   initialize the integral with weight function     */
    template<class _Distribution>
    double assign_integral(_Distribution&&);
    
    /*    initialize the function and the integral    */
    template<class _Distribution, class _Func>
    double assign(_Distribution&&, _Func&&);
    
    
    /*    do the interpolation with distribution   */
    template<class _Distribution>
    double operator()(_Distribution&&, double) const;
    
    /*    interpolate the integral  with distribution
     *      this assumes a pre-call of integral_init(_Distribution&&)
     */
    template<class _Distribution>
    double integral(_Distribution&&, double) const;
    
    
    /*    integral inverse with distribution
     *      this assumes a call of integral_init(_Distribution&&)
     */
    template<class _Distribution>
    double integral_inverse(_Distribution&&, double) const;
    
  private:
    /*    array of gridpoints    */
    std::vector<gridpoint_t> _M_grid;
    
    
    /*   private types   */
    using _Poly = std::array<double, static_cast<std::size_t>(order)+1>;
    
    /*    returns the polynom over the spline   */
    void _M_poly(_Poly& p, std::size_t k) const
    {
      if(spline_type::constant == order) {
        p[0] = std::max(_M_grid[k].func, _M_grid[k+1].func);
      }
      else {
        const double &x1 = _M_grid[k].x, &x2 = _M_grid[k+1].x;
        const double &f1 = _M_grid[k].func, &f2 = _M_grid[k+1].func;
        
        p[0] = (f1*x2-f2*x1)/(x2-x1);
        p[1] = (f2-f1)/(x2-x1);
      }
    }
    
    /*   evaluate the polynomial    */
    double _M_poly(std::size_t k, double x) const
    {
      if(spline_type::constant == order) {
        return std::max(_M_grid[k].func, _M_grid[k+1].func);
      }
      else {
        const double &x1 = _M_grid[k].x, &x2 = _M_grid[k+1].x;
        const double &f1 = _M_grid[k].func, &f2 = _M_grid[k+1].func;
        
        return ((x2-x)*f1 + (x-x1)*f2)/(x2-x1);
      }
    }
  };
  
  
  template<spline_type _Order>
  void spline<_Order>::makegrid(std::size_t n, double xL, double xH)
  {
    /*   resizeing the grid   */
    std::size_t N = std::max<std::size_t>(n+1, 2);
    _M_grid.resize(N);
    
    /*   it is a constant spline, only two points   */
    if(2 == N) {
      _M_grid[0] = gridpoint_t{xL, 0.0, 0.0};
      _M_grid[1] = gridpoint_t{xH, 0.0, 0.0};
      return;
    }
    
    /*   otherwise   */
    double dx = (xH-xL)/n, xT = xL;
    
    std::generate(_M_grid.begin(), _M_grid.end(), [=, &xT]{
      gridpoint_t res{xT, 0.0, 0.0}; xT += dx; return res;
    });
  }
  
  
  template<spline_type _Order>
  void spline<_Order>::makegrid(std::size_t n, std::initializer_list<double> x)
  {
    /*   resizeing the grid   */
    std::size_t N = std::max(n+1, x.size());
    
    /*   when we give explicitly all the grid points, just copy   */
    if(x.size() == N) {
      _M_grid.resize(N);
      std::transform(x.begin(), x.end(), _M_grid.begin(), [=](double xi) { return gridpoint_t{xi, 0.0, 0.0};});
      return;
    }
    
    /*   when we provide only the boundaries   */
    if(x.size() == 2) {
      const double xL = *x.begin(), xH = *(x.end()-1);
      double dx = (xH-xL)/n, xT = xL;
      
      _M_grid.resize(N);
      std::generate(_M_grid.begin(), _M_grid.end(), [=, &xT]{
        gridpoint_t res{xT, 0.0, 0.0}; xT += dx; return res;
      });
      
      return;
    }
    
    /*   otherwise we have to work harder   */
    const double xL = *x.begin(), xH = *(x.end()-1);
    double dx = (xH-xL)/n, sx = 0.0;
    std::size_t Neff = n;
    
    for(auto ix = x.begin()+1; ix != x.end(); ++ix) {
      if(ix[0]-ix[-1] > dx) sx += ix[0]-ix[-1];
      else --Neff;
    }
    
    /*    distrubution the gridpoints piecewise    */
    double dx1 = sx/Neff, xT = xL;
    
    _M_grid.clear();
    _M_grid.reserve(N);
    _M_grid.emplace_back(gridpoint_t{xT, 0.0, 0.0});
    
    for(auto ix = x.begin()+1; ix != x.end(); ++ix)
      if(ix[0]-ix[-1] <= dx1) _M_grid.push_back(gridpoint_t{(xT = ix[0]), 0.0, 0.0});
      else
      {
        auto Ni = static_cast<std::size_t>(std::round((ix[0]-ix[-1])/dx1));
        double dxi = (ix[0]-ix[-1])/Ni;
        
        for(std::size_t k = 1; k <= Ni; ++k)
          _M_grid.push_back(gridpoint_t{(xT += dxi), 0.0, 0.0});
      }
  }
  
  
  template<spline_type _Order> template<class _Dist, class _Func>
  double spline<_Order>::assign(_Dist&& dist, _Func&& func)
  {
    _Poly p;
    std::size_t N = _M_grid.size();
    
    _M_grid[0].func = func(_M_grid[0].x);
    _M_grid[0].ifunc = 0.0;
    
    for(std::size_t k = 1; k < N; ++k)
    {
      _M_grid[k].func = func(_M_grid[k].x);
      
      _M_poly(p, k-1);
      _M_grid[k].ifunc = _M_grid[k-1].ifunc + dist.template integral<order>(p, _M_grid[k-1].x, _M_grid[k].x);
    }
    
    return integral();
  }
  
  template<spline_type _Order> template<class _Dist>
  double spline<_Order>::assign_integral(_Dist&& dist)
  {
    _Poly p;
    std::size_t N = _M_grid.size();
    
    _M_grid[0].ifunc = 0.0;
    
    for(std::size_t k = 1; k < N; ++k) {
      _M_poly(p, k-1);
      _M_grid[k].ifunc = _M_grid[k-1].ifunc + dist.template integral<order>(p, _M_grid[k-1].x, _M_grid[k].x);
    }
    
    return integral();
  }
  
  
  template<spline_type _Order>
  template<class _Distribution>
  double spline<_Order>::operator()(_Distribution&& dist, double x) const
  {
    auto sz = size()-1;
    
    /*   extrapolating by freezing the basepoints  */
    if(x <= _M_grid[0].x) return _M_grid[0].func;
    if(x >= _M_grid[sz].x) return _M_grid[sz].func;
    
    /*   findig the gridpoint  */
    int lo = 0, hi = (int) sz, mi;
    
    while(lo != (mi = (lo+hi)/2)) {
      if(_M_grid[mi].x <= x) lo = mi;
      else hi = mi;
    }
    
    /*   evaluate the polynomial    */
    return dist(x)*_M_poly(lo, x);
  }
  
  
  template<spline_type _Order>
  template<class _Distribution>
  double spline<_Order>::integral(_Distribution&& dist, double x) const
  {
    auto sz = size()-1;
    
    /*  extrapolating by freezing the basepoints  */
    if(x <= _M_grid[0].x) return 0.0;
    if(x >= _M_grid[sz].x) return _M_grid[sz].ifunc + _M_grid[sz].func*(x-_M_grid[sz].x);
    
    /*   findig the gridpoint  */
    int lo = 0, hi = sz, mi;
    
    while(lo != (mi = (lo+hi)/2)) {
      if(_M_grid[mi].x <= x) lo = mi;
      else hi = mi;
    }
    
    /*   calculate the integral   */
    _Poly p;
    _M_poly(p, lo);
    
    return _M_grid[lo].ifunc + dist.template integral<order>(p, _M_grid[lo].x, x);
  }
  
  
  template<spline_type _Order>
  template<class _Distribution>
  double spline<_Order>::integral_inverse(_Distribution&& dist, double F) const
  {
    auto sz = size()-1;
    
    /*   if it is out of range return something reasonable  */
    if(F <= 0.0) return _M_grid[0].x;
    if(F >= _M_grid[sz].ifunc) return _M_grid[sz].x;
    
    /*   findig the gridpoint  */
    int lo = 0, hi = sz, mi;
    
    while(lo != (mi = (lo+hi)/2)) {
      if(_M_grid[mi].ifunc <= F) lo = mi;
      else hi = mi;
    }
    
    /*   calculate the inverse   */
    _Poly p;
    _M_poly(p, lo);
    
    return dist.template integral_inverse<order>(p, _M_grid[lo].x, _M_grid[lo+1].x, F-_M_grid[lo].ifunc);
  }
  
  
  
  /*    This is an adaptor class to the spline<...> class. We can multiply the spline
   *    by a weight function and use it as another spline like object.
   */
  template<class _Dist>
  class weighted_spline_view
  {
  public:
    /*   constructors   */
    template<spline_type _Order, typename ..._Args>
    explicit weighted_spline_view(spline<_Order> *s, _Args ...args)
    : _M_spl{s}, _M_dist{args...}, _M_integral{0.0} {}
    
    template<spline_type _Order, typename ..._Args>
    explicit weighted_spline_view(spline<_Order>& s, _Args ...args)
    : _M_spl{std::addressof(s)}, _M_dist{args...}, _M_integral{0.0} {}
    
    
    /*    Initialize the integral (the underlying spline is already set up!!!)     */
    double assign_integral() const {
      _M_integral = (*_M_spl.assign)(_M_spl.ptr, _M_dist);
      return _M_integral;
    }
    
    /*    do the interpolation with the weight function   */
    double operator()(double x) const {
      return (*_M_spl.func)(_M_spl.ptr, _M_dist, x);
    }
    
    /*    the integral with weight function    */
    double integral(double x) const {
      return (*_M_spl.integral)(_M_spl.ptr, _M_dist, x);
    }
    
    /*    integral over the interpolation range   */
    double integral() const { return _M_integral;}
    
    /*    the integral with weight function    */
    double integral_inverse(double F) const {
      return (*_M_spl.inverse)(_M_spl.ptr, _M_dist, F);
    }
    
  private:
    /*   static function to implement virtual table with type eraser    */
    template<class _Spline>
    struct _Trampoline
    {
      static double func(void *ptr, const _Dist& dist, double x) {
        return reinterpret_cast<_Spline *>(ptr) -> operator()(dist, x);
      }
      
      static double integral(void *ptr, const _Dist& dist, double x) {
        return reinterpret_cast<_Spline *>(ptr) -> integral(dist, x);
      }
      
      static double inverse(void *ptr, const _Dist& dist, double F) {
        return reinterpret_cast<_Spline *>(ptr) -> integral_inverse(dist, F);
      }
      
      static double assign(void *ptr, const _Dist& dist) {
        return reinterpret_cast<_Spline *>(ptr) -> assign_integral(dist);
      }
    };
    
    /*    implements the type erased refernce to a spline    */
    struct _Spline_ref
    {
      /*   constructior from dispatch type    */
      template<class _Spline> _Spline_ref(_Spline *p)
      : ptr{reinterpret_cast<void *>(p)}
      , func{_Trampoline<_Spline>::func}
      , integral{_Trampoline<_Spline>::integral}
      , inverse{_Trampoline<_Spline>::inverse}
      , assign{_Trampoline<_Spline>::assign}
      {}
      
      /*   pointers    */
      void *ptr;
      double (*func)(void *, const _Dist&, double);
      double (*integral)(void *, const _Dist&, double);
      double (*inverse)(void *, const _Dist&, double);
      double (*assign)(void *, const _Dist&);
    } _M_spl;
    
    
    /*   ditribution (should be small object)   */
    _Dist _M_dist;
    
    /*   the integral over the interpolation range    */
    mutable double _M_integral;
  };
  
  
  /*    this is just with unit weight    */
  using spline_view = weighted_spline_view<unitweight>;
  using logspline_view = weighted_spline_view<logweight>;
  
  
}  //   namespace duct


#endif
