/*  Copyright (C) 2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 6/25/13.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __duct_tags_h__
#define __duct_tags_h__ 1



namespace duct {
  
  /*  Empty classes to tag ranges in the class statistical_basis. 
   *  With these we can get some information in a more elegant way.
   */
  struct qcd_out {};
  struct non_qcd {};
  struct incomings {};
  struct outgoings {};

  
  /* Index class to implement such an operator like
   *  event<hhc> p;
   *  lorentzvector<double> pA;
   *
   *  p[hadron(-1)].key = key;
   *  p[hadron(-1)].momentum = pA;
   */
  struct hadron {
    explicit hadron(int k = 0) : idx(k) {}
    int idx;
  };
  
  
  /*  Tagging classes those are related to the initial or final state.
   *  For example: splitting_kernel<ini>
   */
  struct ini {};
  struct fin {};
  
  
  /*  Some simple class to describe the level of some ojects  */
  struct lo {};
  struct nlo {};
  
 
  /*   Class for tagging fixed order or shower calculations   */
  template<class _Level>
  struct shower {
    using level = _Level;
  };
 
  template<class _Level>
  struct fixedorder {
    using level = _Level;
  };
  
  
  /* Tagging the classes those are related to e+e- anniilation. */
  struct epa {
    static constexpr unsigned number_of_incoming_hadrons = 0u;
    static constexpr const char *name = "epa";
  };
  
  /* Tagging the classes those are related to DIS. */
  struct dis {
    static constexpr unsigned number_of_incoming_hadrons = 1u;
    static constexpr const char *name = "dis";
  };
  
  /* Tagging the classes those are related to hadron-hadron collision. */
  struct hhc {
    static constexpr unsigned number_of_incoming_hadrons = 2u;
    static constexpr const char *name = "hhc";
  };
  
  /* Tagging the classes those are related to heavy particle decay. */
  struct decay {
    static constexpr unsigned number_of_incoming_hadrons = 0u;
    static constexpr const char *name = "decay";
  };
  
    
  /*   ordering scheme   */
  struct ordering
  {
    /*   alpha parameter of the ordering   */
    enum class scheme : unsigned int { kT = 0u, lambda = 1u, angular = 2u};
    
    /*  Simple class for tagging objects for kT ordering   */
    struct kT {
      static constexpr const char *name = "kT";
      static constexpr scheme alpha = scheme::kT;
    };
    
    /*  Simple class for tagging objects for lambda ordering   */
    struct lambda {
      static constexpr const char *name = "lambda";
      static constexpr scheme alpha = scheme::lambda;
    };
    
    /*  Simple class for tagging objects for kT ordering   */
    struct angular {
      static constexpr const char *name = "angular";
      static constexpr scheme alpha = scheme::angular;
    };
  };
  
  
  /*   Some functions depends on the splitting type and we have to provide an agrument
   *   and switch behaviour according to this argument. So, we have a simple enumm class to 
   *   to represent the three cases. 
   *   soft                      => soft part of the f -> f+g vertex
   *   collinear                 => collinear part of the f -> f+g vertex
   *   flavor_changing_collinear => flavour changing vertices
   */
  enum class splitting_type {
    soft = 0,
    collinear = 1,
    flavor_changing_collinear = 2
  };

  
  /*   In this code everything is expressed in GeV,
   *   as input we might want to give some quantities in TeV, MeV.
   */
  constexpr double operator"" _MeV(long double s) { return s/1000.0;}
  constexpr double operator"" _GeV(long double s) { return s;}
  constexpr double operator"" _TeV(long double s) { return 1000.0*s;}
  
  constexpr double operator"" _MeV(unsigned long long s) { return s/1000.0;}
  constexpr double operator"" _GeV(unsigned long long s) { return s;}
  constexpr double operator"" _TeV(unsigned long long s) { return 1000.0*s;}
 
  
  /*   check if the type _Ordering is ordering tag   */
  template<class _Ordering>
  constexpr bool is_ordering = (std::is_same<_Ordering, ordering::kT>::value
                                || std::is_same<_Ordering, ordering::lambda>::value
                                || std::is_same<_Ordering, ordering::angular>::value);
  
  
  /*   chack if there is incoming hadron   */
  template<class _Proc>
  constexpr bool is_hadron_collision = std::is_same<_Proc, dis>::value || std::is_same<_Proc, hhc>::value;
}

#endif
