/*  Copyright (C) 2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 3/13/13.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __mc_splitfunc_h__
#define __mc_splitfunc_h__ 1

/*   std includes   */
#include <array>

/*   project headers   */
#include <bits/mc-splitfunc-base.h>
#include <bits/mc-splitfunc-approx.h>



namespace duct {

  
  template<class _Ordering>
  struct splitting_functions
  {
    /*    public types    */
    using ordering_type = _Ordering;
    
    /*   The initial state splitting functions     */
    class initial_state
    {
      /*    type aliases   */
      using _Base = __details::splitting_functions_base<ini>;
      using _EmitG = __details::splitting_functions_base<ini>::emitting_gluon;

    public:
      /*    class for representing the soft state   */
      class soft_state
      {
      public:
        /*    return the index of the spectator parton   */
        constexpr int spectator() const { return _M_spectator;}
        
        /*   gets full weight with the color factor   */
        constexpr double weight() const { return _M_weight;}
        
        /*   gets the azimuthal angle from a random number   */
        double azimuthal_angle(double rho) const { return _M_state.inverse_phi(rho);}
        
      private:
        /*    set the state   */
        double set(const hard_params<ini>::spectator_vars& vars, double y, double z, double fac)
        {
          _M_weight = _M_state(vars.xi, vars.xT, y, z);
          _M_weight *= fac*vars.cfac;
          _M_spectator = vars.spectator;
          return _M_weight;
        }
        
        /*   private data members   */
        int _M_spectator;
        double _M_weight;
        _EmitG _M_state;
        
        /*   initial state class is a friend    */
        friend class initial_state;
      };
      
      /*    calculating the spilitting function   */
      double operator()(splitting_type, const hard_params<ini> *, double, double);
      
      /*   get the pdf ratio   */
      double pdf_ratio(int k) const { return _M_pdf_ratio[k+5];}
      
      /*   returns the soft states   */
      const stackvector<soft_state, 4>& soft_states() const { return _M_soft;}
      
      /*    Defines the vetos. This is needed when the hard part is not uniquely defined
       *    like in process p + p -> jets.
       */
      static bool veto(const hard_params<ini> *pars, double v, double z)
      {
        double y = ordering_traits<ini,_Ordering>::mapping_to_y_nojac(pars, v, z);
        return ((1.0-z)*y < pars->mTmax && y < pars->mYmax);
      }

    private:
      /*   stores the soft gluon states for the possible spectators   */
      stackvector<soft_state, 4> _M_soft;
      
      /*    array to store the pdf ration when the emitter is gluon and it turns to a quark    */
      std::array<double,11> _M_pdf_ratio;
    };
    
    
    /*   The final state splitting functions     */
    class final_state
    {
      /*    type aliases   */
      using _Base = __details::splitting_functions_base<fin>;
      using _EmitG = __details::splitting_functions_base<fin>::emitting_gluon;
      
    public:
      /*    class for representing the soft state   */
      class soft_state
      {
      public:
        /*    return the index of the spectator parton   */
        int spectator() const { return _M_spectator;}
        
        /*   gets full weight with the color factor   */
        double weight() const { return _M_weight;}
        
        /*   gets the azimuthal angle from a random number   */
        double azimuthal_angle(double rho) const { return _M_state.inverse_phi(rho);}
        
      private:
        /*    set the state   */
        double set(const hard_params<fin> *pars, const hard_params<fin>::spectator_vars& vars, double y, double z, double fac)
        {
          _M_weight = _M_state(pars, vars, y, z);
          _M_weight *= fac*vars.cfac;
          _M_spectator = vars.spectator;
          return _M_weight;
        }
        
        /*   private data members   */
        int _M_spectator;
        double _M_weight;
        _EmitG _M_state;
        
        /*   initial state class is a friend    */
        friend class final_state;
      };
      
      /*   calculating the soft part   */
      double operator()(splitting_type, const hard_params<fin> *, double, double);
    
      /*   returns the soft states   */
      const stackvector<soft_state, 4>& soft_states() const { return _M_soft;}
      
      /*   returns the collinear probabilities   */
      const stackvector<double, 7>& collinear_weights() const { return _M_coll;}
    
      /*    Defines the vetos. This is needed when the hard part is not uniquely defined
       *    like in process p + p -> jets.
       */
      static bool veto(const hard_params<fin> *pars, double v, double z)
      {
        double y = ordering_traits<fin,_Ordering>::mapping_to_y_nojac(pars, v, z);
        return (z*(1.0-z)*y*y/(y+pars->bl) < pars->mTmax && y < pars->mYmax);
      }
      
    private:
      /*   stores the soft gluon states for the possible spectators   */
      stackvector<soft_state, 4> _M_soft;
 
      /*   in the collinear splitting this array gives the probabilities of each channels    */
      stackvector<double, 7> _M_coll;
   };
    
    
    /*   The approximation of the splitting function. This classes can do the mapping
     *   from uniform random numbers to the splitting variables.
     */
    struct approx {
      using initial_state = __details::splitting_functions_approx<ini, _Ordering>;
      using final_state = __details::splitting_functions_approx<fin, _Ordering>;
    };
  };

  
  template<class _Ordering>
  double splitting_functions<_Ordering>::initial_state::
  operator()(splitting_type idx, const hard_params<ini> *pars, double v, double z)
  {
    /*   transform variables from (v,z) to (y,z)    */
    auto __y = ordering_traits<ini,_Ordering>::mapping_to_y(pars, v, z);
    const double &y = __y.first, &jac = __y.second;
        
    /*    calculate the splitting functions   */
    switch(idx)
    {
      case splitting_type::soft:
      {
        /*   make sopace for the soft states   */
        _M_soft.resize(pars->spectators.size());
        
        /*    loop over the possible spectators   */
        double ret = 0.0;
        unsigned k = 0u;
        
        for(const auto& spec : pars->spectators)
          ret += _M_soft[k++].set(spec, y, z, jac);
        
        return ret;
      } break;
        
      case splitting_type::collinear:
        return jac*(pars->fl ? _Base::quark_gluon(y, z) : _Base::gluon_gluon(y, z));
        break;
        
      case splitting_type::flavor_changing_collinear:
      {
        /*   quark emitter    */
        if(pars->fl) return jac*_Base::quark_antiquark(y, z);
        
        /*   gluon emitter    */
        else {
          double singlet = 0.0;
          
          for(int k = -3; k <= 5; ++k)
            if(k) singlet += (_M_pdf_ratio[k+5] = pars->aux->parton(-k, pars->eta/z, pars->Q2*v));
          
          singlet += (_M_pdf_ratio[0] = _M_pdf_ratio[10]);
          singlet += (_M_pdf_ratio[1] = _M_pdf_ratio[9]);
          _M_pdf_ratio[5] = 0.0;
          
          for(double& r : _M_pdf_ratio)
            r /= singlet;
          
          return jac*_Base::gluon_quark(y, z);
        }
      } break;
    }
    
    return 0.0;
  }

  
  template<class _Ordering>
  double splitting_functions<_Ordering>::final_state::
  operator()(splitting_type idx, const hard_params<fin> *pars, double v, double z)
  {
    /*   transform variables from (v,z) to (y,z)    */
    auto __y = ordering_traits<fin,_Ordering>::mapping_to_y(pars, v, z);
    const double &y = __y.first, &jac = __y.second;
    
    /*    calculate the splitting functions   */
    switch(idx) {
      case splitting_type::soft:
      {
        /*   make sopace for the soft states   */
        _M_soft.resize(pars->spectators.size());
        
        /*    loop over the possible spectators   */
        double ret = 0.0;
        unsigned k = 0u;
        
        for(const auto& spec : pars->spectators)
          ret += _M_soft[k++].set(pars, spec, y, z, jac);
        
        return ret;
      } break;
        
      case splitting_type::collinear:
      {
        /*    quark emitter   */
        if(pars->fl)
        {
          _M_coll.resize(1u);
          _M_coll[0] = jac*_Base::quark_gluon(pars, y, z);
          return _M_coll[0];
        }
 
        /*   gluon emitter    */
        else {
          _M_coll.resize(7u);
 
          /*   g -> g + g  (pure collinear)    */
          _M_coll[0] = jac*_Base::gluon_gluon(pars, y, z);
          
          /*   g -> q + q~      */
          double norm = _M_coll[0], wqq = jac*_Base::quark_antiquark(81, pars, y, z);
          
          /*    we have five light flavors in this splitting...    */
          for(int k = 1; k < 6; ++k) norm += (_M_coll[k] = wqq);
          
          /*   and we only consider the top quark contribution in g -> Q + Q~ vertex   */
          norm += (_M_coll[6] = jac*_Base::quark_antiquark(6, pars, y, z));
          
          return norm;
        }
      } break;
        
      default:
        return 0.0;
    }
  }

  
  
  
}   /*   namespace duct    */


#endif

