/*  Copyright (C) 2011-2013 Zoltán Nagy
 *  Created by Zoltan Nagy on 13/5/11.
 *  
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>. 
 */
#ifndef __duct_qcd_color_h__
#define __duct_qcd_color_h__ 1


/*   std headers   */
#include <map>
#include <string>
#include <functional>

/*   project headers    */
#include <bits/duct-tags.h>


namespace duct {
  namespace qcd {
 
    /*   enum class for tagging the order in alpha_s   */
    enum class order : unsigned { lo = 0, nlo = 1};
      
      
    /*   Simple structure to provide QCD imput.
     *   It makes the user module simpler and clearer (I hope).
     */
    struct input
    {
      /*  name of the qcd input   */
      std::string name = "default";
      
      /*  order of the strong couling   */
      qcd::order order = qcd::order::nlo;
      
      /*   coupling at the reference scale   */
      double qref = 91.1876_GeV, asref = 0.118;
      
      /*   heavy flavor thresholds     */
      double mc = 1.3_GeV, mb = 4.5_GeV, mt = 1e+100;
      
      /*   pdf fitting scale (transverse momentum cutoff)    */
      double mT = 1.0_GeV;
      
      /*    pdf at the fitting scale    */
      std::map<int, std::function<void(double, double*)>> pdf;
    };
    
    
    
    /*     color factors (SU(N) and Tr = 0.5)    */
    constexpr double Tr  = 0.5;                         //   Tr  = 0.5
    constexpr double Nc  = 3.0;                         //   Nc  = 3;
    constexpr double Na  = 8.0;                         //   Na  = Nc^2-1;
    constexpr double Nc2 = Nc*Nc;                       //   Nc2 = Nc^2;
    constexpr double Na2 = Na*Na;                       //   Na2 = Na^2;
    constexpr double Cf  = 0.5*Na/Nc;                   //   Cf  = Na/(2*Nc)
    constexpr double Ca  = Nc;                          //   Ca  = Nc;
    
    
    /*   beta function coefficients   */
    constexpr double beta1(unsigned nf) {
      return (11.0*Ca - 2.0*nf)/6.0;
    }
    
    constexpr double beta2(unsigned nf) {
      return (17.0*Ca*Ca - 3.0*Cf*nf - 5.0*Ca*nf)/6.0;
    }
    
    
    /*  gamma parameters in the Altarelli-Parisi functions   */
    constexpr double G1q  = 1.5*Cf;
    
    constexpr double G1g(unsigned nf) {
      return (11.0*Ca - 2.0*nf)/6.0;
    }
    
    constexpr double G2q(unsigned nf)
    {
      constexpr double zeta2 = 1.64493406684822643647241516665;
      constexpr double zeta3 = 1.20205690315959428539973816151;
      
      constexpr double p1 = 17.0/24.0 + 11.0/3.0*zeta2 - 3.0*zeta3;
      constexpr double p2 = 1.0/12.0 + 2.0/3.0*zeta2;
      constexpr double p3 = 3.0/8.0 - 3.0*zeta2 + 6.0*zeta3;
      
      return Cf*(Ca*p1 - nf*p2 + Cf*p3);
    }
    
    constexpr double G2g(unsigned nf)
    {
      constexpr double zeta3 = 1.20205690315959428539973816151;
      constexpr double p1 = -2.0/3.0;
      constexpr double p2 = 8.0/3.0+ 3.0*zeta3;
      constexpr double p3 = - 0.5;
      
      return Ca*(nf*p1  + Ca*p2) + Cf*nf*p3;
    }
    
    
    /*   Kg = Ca*(67/18-zeta2)-10*Tr*nf/9    */
    constexpr double Kg(unsigned nf) {
      return 2.07728815537399578575*Ca - 0.55555555555555555555*nf;
    }
  }   //  namespace qcd
}   //   namespace duct
#endif
