/*  Copyright (C) 2014 Zoltán Nagy
 *  Created by Zoltán Nagy on 2/6/14.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

/*   std includes   */

/*   project headers  */
#include "bits/qcd-color.h"
#include "bits/mc-splitting-prob.h"
#include "cxx/math-support.h"


namespace duct {
  
  namespace __details {
    
    void counting_color_connections(stackvector<std::pair<int,double>, 4>& spectator, int emitter, const particle_state *p, const model_ref& mdl)
    {
      spectator.clear();
      
      switch(mdl.color(p[emitter].flavor))
      {
        case 8:
        {
          double cfac = 0.25*qcd::Ca;
         
          spectator.push_back({p[emitter].color.prev, cfac});
          spectator.push_back({p[emitter].color.next, cfac});
          
          if(p[emitter].colorP.prev == spectator[0].first) spectator[0].second += cfac;
          else if(p[emitter].colorP.prev == spectator[1].first) spectator[1].second += cfac;
          else spectator.push_back({p[emitter].colorP.prev, cfac});
          
          if(p[emitter].colorP.next == spectator[0].first) spectator[0].second += cfac;
          else if(p[emitter].colorP.next == spectator[1].first) spectator[1].second += cfac;
          else spectator.push_back({p[emitter].colorP.next, cfac});
        } break;
          
        case 3:
        {
          double cfac = 0.5*qcd::Cf;
          
          spectator.push_back({p[emitter].color.next, cfac});
          
          if(p[emitter].colorP.next == spectator[0].first) spectator[0].second += cfac;
          else spectator.push_back({p[emitter].colorP.next, cfac});
        } break;
        
        case -3:
        {
          double cfac = 0.5*qcd::Cf;
        
          spectator.push_back({p[emitter].color.prev, cfac});
          
          if(p[emitter].colorP.prev == spectator[0].first) spectator[0].second += cfac;
          else spectator.push_back({p[emitter].colorP.prev, cfac});
        } break;
      }
    }
    
    
    
    /*    Gauss integrals    */
    const gauss_integral<10u> splitting_prob_base<true>::threshold_integralT{};
    static const gauss_integral<10u> _threshold_integralZ{};
    
    
    /*   the full soft function    */
    static double __parton_gluon_sum(const shower_engine_opts::threshold_opts& opts,  const hard_params<ini> *pars, double y, double z)
    {
      /*   spectator independent quantities   */
      double retval = 0.0, oz = 1.0-z;
      
      /*   loop over spectators   */
      for(const auto& spec : pars->spectators)
        if(spec.spectator > 0 && spec.xi > 1.0e-6)
        {
          /*   Dave's variable for angle. I think 1.0-spec.xi == 0.5*(1-cos(theta)) would be simpler....  */
          double psi = 0.5*(1.0-spec.xi)/std::sqrt(spec.xi);
          
          if(opts.dynamic_psi_cut ? oz*psi > z*y : psi > opts.psi_min)
            retval += 2.0*spec.cfac*(1.0/oz - 1.0/std::sqrt(sqr(oz) + sqr(y*z/psi)));
        }
      
      return retval;
    }
    
    
    
    /*   Functions for calculating the threshold weights for kT ordering.    */
    template<>
    thrs_weight splitting_prob_base<true>::
    threshold<ordering::kT>(const shower_engine_opts::threshold_opts& opts, const hard_params<ini> *pars, double tau)
    {
      const qcd_coupling& as = pars->mdl.alpha_qcd();
      double q2 = as.q2(tau);
      double v = q2/pars->Q2;
      
      /*   the z limits are scheme dependent    */
      double zmin = 4.0/sqr(std::sqrt(v) + std::sqrt(4.0+v));
      double zmax = (1.0-pars->eta)*0.999 + pars->eta;
      
      /*   defining the integrands of the z integrals    */
      int fa = -pars->fl;
      double pdf = pars->aux->parton(fa, pars->eta, q2);
      
      
      /*   integrand of the main term    */
      auto color_diag = [=](double z) -> std::pair<double, double> {
        /*   When z < eta then f(eta/z, q2) = 0 for every flavors.
         *   In this case the integrand is simple.
         */
        double oz = 1.0-z;
        
        if(z < pars->eta)
          return {(fa ? 2.0*qcd::Cf : 2.0*qcd::Ca)/oz, 0.0};
        
        
        /*   Otherwise we have a more complicated function   */
        double reg = 0.0, ratio, soft = (fa ? 2.0*qcd::Cf : 2.0*qcd::Ca);
        
        /*   soft term    */
        if(z >= 0.9999) {
          ratio = pars->aux->parton(fa, pars->eta/0.9999, q2)/pdf;
          soft *= (1.0 - ratio*z)/0.0001;
        }
        else {
          ratio = pars->aux->parton(fa, pars->eta/z, q2)/pdf;
          soft *= (1.0 - ratio*z)/oz;
        }
        
        
        /*   collinear terms : quark emitter   */
        if(fa) {
          reg  = qcd::Tr*(1.0-2.0*z*oz);
          reg *= pars->aux->parton(0, pars->eta/z, q2)/pdf;
          reg += qcd::Cf*oz*ratio;
        }
        
        /*   collinear terms : gluon emitter   */
        else {
          reg = qcd::Cf*(z + 2.0*oz/z);
          reg *= pars->aux->parton(81, pars->eta/z, q2)/pdf;
          reg += 2.0*qcd::Ca*(oz/z+z*oz)*ratio;
       }
        
        return {soft, -reg};
      };
      
      
      /*    jacobians    */
      double Js = 0.5*(1.0-zmax);
      double Jm = 0.5*std::log((1.0-zmin)/(1.0-zmax));
      double Jd = 0.5*std::log(1.0/(1.0-zmin));

      return _threshold_integralZ([=](double xi) -> thrs_weight {
        /*   color diagonal integrand close to z=1   */
        double soft, reg;
        //------
        {
          double z = Js*(1.0+xi) + zmax;
          auto d = color_diag(z);
          
          soft = Js*d.first;
          reg = Js*d.second;
        }
        
        
        /*    color diagonal integral    */
        if(zmax > zmin)
        {
          double z = 1.0-(1.0-zmin)*std::exp(-(1.0+xi)*Jm);
          double oz = 1.0-z;
          auto d = color_diag(z);
          
          soft += oz*Jm*d.first;
          reg += oz*Jm*d.second;
        }
        
        
        /*    color off diagonal integral    */
        double delta;
        //------
        {
          /*   integral transformation    */
          double z = 1.0-1.0*std::exp(-(1.0+xi)*Jd);
          double oz = 1.0-z;
          
          /*    sum over the specators  (only LC+ connections)    */
          delta = __parton_gluon_sum(opts, pars, v/oz, z);
          
          /*   PDF factor    */
          if(z > pars->eta)
            delta *= 1.0 - pars->aux->parton(fa, pars->eta/z, q2)*z/pdf;
          
          delta *= oz*Jd;
        }
        
  
        return {soft, reg, delta};
      });
    }


    
    /*   Functions for calculating the threshold weights for lambda ordering.    */
    template<>
    thrs_weight splitting_prob_base<true>::
    threshold<ordering::lambda>(const shower_engine_opts::threshold_opts& opts, const hard_params<ini> *pars, double tau)
    {
      const qcd_coupling& as = pars->mdl.alpha_qcd();
      
      double q2 = as.q2(tau);
      double v = q2/pars->Q2;
      
      /*   the z limits are scheme dependent    */
      double zmin = 1.0/(1.0+v);
      double zmax = 1.0-pars->mT/v;
      double umin = std::max(0.0, 1.0-pars->mTmax/v);
      double umax = std::min(zmin, zmax);
      
      if(zmin < umin) zmin = umin;
      
      
      /*   defining the integrands of the z integrals    */
      int fa = -pars->fl;
      double pdf = pars->aux->parton(fa, pars->eta, q2);
      double asb = as.ask(tau)*qcd::beta1(as.nf(tau));
      
      /*     jacobians    */
      double Jm = 0.5*std::log((1.0-zmin)/(1.0-zmax));
      double Jd = 0.5*std::log((1.0-umin)/(1.0-umax));
 
      
      return _threshold_integralZ([=](double xi) -> thrs_weight {
        /*    color diagonal integrand    */
        double soft = 0.0, reg = 0.0;
        
        if(zmax > zmin)
        {
          /*   integral transformation    */
          double z = 1.0-(1.0-zmin)*std::exp(-(1.0+xi)*Jm);
          double oz = 1.0-z;
          
          /*   When z < eta then f(eta/z, q2) = 0 for every flavors.
           *   In this case the integrand is simple.
           */
          if(z < pars->eta) {
            soft  = Jm*(fa ? 2.0*qcd::Cf : 2.0*qcd::Ca);
            soft /= 1.0+asb*std::log(oz);
          }
          
          /*   Otherwise we have a more complicated function   */
          else {
            double ratio = pars->aux->parton(fa, pars->eta/z, q2)/pdf;
            
            /*   soft term    */
            soft = Jm*(1.0 - ratio*z)*(fa ? 2.0*qcd::Cf : 2.0*qcd::Ca);
            soft /= 1.0+asb*std::log(oz);
            
            /*   quark emitter   */
            if(fa) {
              reg  = qcd::Tr*(1.0-2.0*z*oz);
              reg *= pars->aux->parton(0, pars->eta/z, q2)/pdf;
              reg += qcd::Cf*oz*ratio;
            }
            
            /*   gluon emitter   */
            else {
              reg  = qcd::Cf*(z + 2.0*oz/z);
              reg *= pars->aux->parton(81, pars->eta/z, q2)/pdf;
              reg += 2.0*qcd::Ca*(oz/z+z*oz)*ratio;
            }
          
            reg *= -Jm*oz;
          }
        }
        
        
        /*    color off diagonal integrand    */
        double delta = 0.0;
        
        if(umax > umin)
        {
          /*   integral transformation    */
          double z = 1.0-(1.0-umin)*std::exp(-(1.0+xi)*Jd);
          double oz = 1.0-z;
          
          /*    sum over the specators  (only LC+ connections)    */
          delta = __parton_gluon_sum(opts, pars, v, z);
          
          /*   PDF factor   */
          if(z > pars->eta)
            delta *= 1.0 - pars->aux->parton(fa, pars->eta/z, q2)*z/pdf;
          
          delta /= 1.0+asb*std::log(oz);
          delta *= oz*Jd;
        }
        
        return {soft, reg, delta};
      });
    }
   
    
    
    /*   Functions for calculating the threshold weights for angular ordering.    */
    template<>
    thrs_weight splitting_prob_base<true>::
    threshold<ordering::angular>(const shower_engine_opts::threshold_opts& opts, const hard_params<ini> *pars, double tau)
    {
      const qcd_coupling& as = pars->mdl.alpha_qcd();
      double q2 = as.q2(tau);
      double zmin = 0.0, zmax = 1.0, umin = 0.0, umax= 0.99999, v = q2/pars->Q2;
      
      
      /*   the z limits are scheme dependent    */
      zmin = std::min(1.0, 1.0/v);
      zmax = 1.0-std::sqrt(pars->mT/v);
      umin = std::max(0.0, 1.0-std::sqrt(pars->mTmax/v));
      umax = std::min(zmin, zmax);
      if(zmin < umin) zmin = umin;
      
      
      /*   defining the integrands of the z integrals    */
      int fa = -pars->fl;
      double pdf = pars->aux->parton(fa, pars->eta, q2);
      double asb = 2.0*as.ask(tau)*qcd::beta1(as.nf(tau));
      
  
      /*     jacobians    */
      double Jm = 0.5*std::log((1.0-zmin)/(1.0-zmax));
      double Jd = 0.5*std::log((1.0-umin)/(1.0-umax));
      
      
      return _threshold_integralZ([=](double xi) -> thrs_weight {
        /*    color diagonal integrand    */
        double soft = 0.0, reg = 0.0;
        
        if(zmax > zmin)
        {
          /*   integral transformation    */
          double z = 1.0-(1.0-zmin)*std::exp(-(1.0+xi)*Jm);
          double oz = 1.0-z;
          
          /*   When z < eta then f(eta/z, q2) = 0 for every flavors.
           *   In this case the integrand is simple.
           */
          if(z < pars->eta) {
            reg = 0.0;
            soft = (fa ? 2.0*qcd::Cf : 2.0*qcd::Ca);
            soft *= Jm/(1.0+asb*std::log(oz));
          }
          
          /*   Otherwise we have a more complicated function   */
          else {
            double ratio = pars->aux->parton(fa, pars->eta/z, q2)/pdf;
            
            /*   soft term    */
            soft = Jm*(1.0 - ratio*z)*(fa ? 2.0*qcd::Cf : 2.0*qcd::Ca);
            soft /= 1.0+asb*std::log(oz);
            
            /*   quark emitter   */
            if(fa) {
              reg  = qcd::Tr*(1.0-2.0*z*oz);
              reg *= pars->aux->parton(0, pars->eta/z, q2)/pdf;
              reg += qcd::Cf*oz*ratio;
            }
            
            /*   gluon emitter   */
            else {
              reg  = qcd::Cf*(z + 2.0*oz/z);
              reg *= pars->aux->parton(81, pars->eta/z, q2)/pdf;
              reg += 2.0*qcd::Ca*(oz/z+z*oz)*ratio;
            }
            
            reg *= -Jm*oz;
          }
        }
        
        
        /*    color off diagonal integrand    */
        double delta = 0.0;
        
        if(umax > umin)
        {
          /*   integral transformation    */
          double z = 1.0-(1.0-umin)*std::exp(-(1.0+xi)*Jd);
          double oz = 1.0-z;
          
          /*    sum over the specators  (only LC+ connections)    */
          delta = __parton_gluon_sum(opts, pars, oz*v, z);
          
          /*   PDF factor   */
          if(z > pars->eta)
            delta *= 1.0 - pars->aux->parton(fa, pars->eta/z, q2)*z/pdf;
          
          delta /= 1.0+asb*std::log(oz);
          delta *= oz*Jd;
        }
        
        return {soft, reg, delta};
      });
    }
    
    
  }   /*   namespace __details   */


}

